from schema_registry.client import errors, schema  # noqa
from schema_registry.client.client import AsyncSchemaRegistryClient, SchemaRegistryClient  # noqa

__all__ = ["SchemaRegistryClient", "AsyncSchemaRegistryClient"]



import typing
from collections.abc import Mapping, Sequence

from schema_registry.client import SchemaRegistryClient
from schema_registry.client.schema import AvroSchema, BaseSchema, JsonSchema
from schema_registry.serializers import AvroMessageSerializer, JsonMessageSerializer, MessageSerializer

try:
    from faust import Codec, Record
except ImportError as ex:
    raise Exception("Cannot use Faust serializers Faust is not installed.") from ex


class Serializer(Codec):
    def __init__(
        self,
        schema_subject: str,
        schema: typing.Union[BaseSchema],
        message_serializer: MessageSerializer,
    ):
        self.schema_subject = schema_subject
        self.schema = schema
        self.message_serializer = message_serializer

        Codec.__init__(self)

    def _loads(self, event: bytes) -> typing.Optional[typing.Dict]:

        return self.message_serializer.decode_message(event)

    def _dumps(self, payload: typing.Dict[str, typing.Any]) -> bytes:
        """
        Given a parsed avro schema, encode a record for the given topic.  The
        record is expected to be a dictionary.
        The schema is registered with the subject of 'topic-value'
        """
        payload = self.clean_payload(payload)

        return self.message_serializer.encode_record_with_schema(
            self.schema_subject, self.schema, payload
        )  # type: ignore

    @staticmethod
    def _clean_item(item: typing.Any) -> typing.Any:
        if isinstance(item, Record):
            return Serializer._clean_item(item.to_representation())
        elif isinstance(item, str):
            # str is also a sequence, need to make sure we don't iterate over it.
            return item
        elif isinstance(item, Mapping):
            return type(item)({key: Serializer._clean_item(value) for key, value in item.items()})  # type: ignore
        elif isinstance(item, Sequence):
            return type(item)(Serializer._clean_item(value) for value in item)  # type: ignore
        return item

    @staticmethod
    def clean_payload(payload: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
        """
        Try to clean payload retrieve by faust.Record.to_representation.
        All values inside payload should be native types and not faust.Record
        On Faust versions <=1.9.0 Record.to_representation always returns a dict with native types
        as a values which are compatible with fastavro.
        On Faust 1.10.0 <= versions Record.to_representation always returns a dic but values
        can also be faust.Record, so fastavro is incapable of serialize them
        Args:
            payload (dict): Payload to clean
        Returns:
            dict that represents the clean payload
        """
        return Serializer._clean_item(payload)


def avro_serializer_factory(
    schema_registry_client: SchemaRegistryClient,
    schema_subject: str,
    schema: AvroSchema,
    return_record_name: bool = False,
) -> "Serializer":  # type: ignore # noqa: F821

    if isinstance(schema, str):
        schema = AvroSchema(schema)

    return Serializer(
        schema_subject, schema, AvroMessageSerializer(schema_registry_client, return_record_name=return_record_name)
    )


def json_serializer_factory(
    schema_registry_client: SchemaRegistryClient,
    schema_subject: str,
    schema: JsonSchema,
    return_record_name: bool = False,
) -> "Serializer":  # type: ignore # noqa: F821

    if isinstance(schema, str):
        schema = JsonSchema(schema)

    return Serializer(
        schema_subject, schema, JsonMessageSerializer(schema_registry_client, return_record_name=return_record_name)
    )


FaustSerializer = avro_serializer_factory
FaustJsonSerializer = json_serializer_factory



from __future__ import annotations
from abc import ABC, abstractmethod
from schema_registry.client.utils import AVRO_SCHEMA_TYPE, JSON_SCHEMA_TYPE

from dataclasses import dataclass

import json
import typing

import aiofiles
import fastavro
import jsonschema


class BaseSchema(ABC):
    def __init__(self, schema: typing.Union[str, typing.Dict]) -> None:
        if isinstance(schema, str):
            schema = json.loads(schema)
        self.raw_schema = schema
        self.schema = self.parse_schema(typing.cast(typing.Dict, schema))
        self.generate_hash()

    @abstractmethod
    def parse_schema(self, schema: typing.Dict) -> typing.Dict:
        pass

    @staticmethod
    @abstractmethod
    def load(fp: str) -> BaseSchema:
        """Parse a schema from a file path"""
        pass

    @staticmethod
    @abstractmethod
    async def async_load(fp: str) -> BaseSchema:
        """Parse a schema from a file path"""
        pass

    @property
    @abstractmethod
    def name(self) -> typing.Optional[str]:
        pass

    @property
    @abstractmethod
    def schema_type(self) -> str:
        pass

    def generate_hash(self) -> None:
        self._hash = hash(json.dumps(self.schema))

    def __hash__(self) -> int:
        return self._hash

    def __str__(self) -> str:
        return str(self.schema)

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, BaseSchema):
            return NotImplemented
        return self.__hash__() == other.__hash__()


class AvroSchema(BaseSchema):
    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self._expanded_schema: typing.Optional[str] = None
        self._flat_schema: typing.Optional[str] = None

        super().__init__(*args, **kwargs)

    @property
    def name(self) -> typing.Optional[str]:
        return self.schema.get("name")

    @property
    def schema_type(self) -> str:
        return AVRO_SCHEMA_TYPE

    @property
    def expanded_schema(self) -> str:
        """
        Returns a schema where all named types are expanded to their real schema

        Returns:
            expanced_schema (str): Schema parsed expanded
        """
        if self._expanded_schema is None:
            self._expanded_schema = fastavro.schema.expand_schema(self.raw_schema)
        return self._expanded_schema

    @property
    def flat_schema(self) -> str:
        """
        Parse the schema removing the fastavro write_hint flag __fastavro_parsed

        Returns:
            flat_schema (str): Schema parsed without the write hint
        """
        if self._flat_schema is None:
            self._flat_schema = fastavro.parse_schema(self.raw_schema, _write_hint=False, _force=True)

        return self._flat_schema

    def parse_schema(self, schema: typing.Dict) -> typing.Dict:
        return fastavro.parse_schema(schema, _force=True)

    @staticmethod
    def load(fp: str) -> AvroSchema:
        """Parse an avro schema from a file path"""
        with open(fp, mode="r") as f:
            content = f.read()
            return AvroSchema(content)

    @staticmethod
    async def async_load(fp: str) -> AvroSchema:
        """Parse an avro schema from a file path"""
        async with aiofiles.open(fp, mode="r") as f:
            content = await f.read()
            return AvroSchema(content)


class JsonSchema(BaseSchema):
    @property
    def name(self) -> typing.Optional[str]:
        return self.schema.get("title", self.schema.get("$id", self.schema.get("$ref")))

    @property
    def schema_type(self) -> str:
        return JSON_SCHEMA_TYPE

    def parse_schema(self, schema: typing.Dict) -> typing.Dict:
        jsonschema.Draft7Validator.check_schema(schema)
        return schema

    @staticmethod
    def load(fp: str) -> BaseSchema:
        """Parse a json schema from a file path"""
        with open(fp, mode="r") as f:
            content = f.read()
            return JsonSchema(content)

    @staticmethod
    async def async_load(fp: str) -> BaseSchema:
        """Parse a json schema from a file path"""
        async with aiofiles.open(fp, mode="r") as f:
            content = await f.read()
            return JsonSchema(content)


class SchemaFactory:
    @staticmethod
    def create_schema(schema: str, schema_type: str) -> typing.Union[JsonSchema, AvroSchema]:
        if schema_type == JSON_SCHEMA_TYPE:
            return JsonSchema(schema)
        elif schema_type == AVRO_SCHEMA_TYPE:
            return AvroSchema(schema)
        else:
            raise ValueError(f"Unsupported schema type '{schema_type}'. Supported schemas are 'AVRO' and 'JSON'.")


@dataclass
class SubjectVersion(object):
    subject: str
    version: int



import typing
import urllib
from collections import defaultdict


class UrlManager:
    def __init__(self, base_url: str, paths: list) -> None:
        parsed_url = urllib.parse.urlparse(base_url)

        assert parsed_url.scheme, f"The url does not have a schema, add one. For example http://{base_url}"

        # this is the absolute url to the server
        # make sure that url ends with /
        if not base_url.endswith("/"):
            base_url += "/"

        self.base_url = base_url

        self.paths = {path.name: path for path in map(lambda path: Path(path), paths)}

    @property
    def url(self) -> str:
        return self.base_url

    def url_for(self, func: str, **kwargs: typing.Any) -> tuple:
        """
        Generate a url for a given function
        """
        path = self.paths[func]
        url = path.generate_url(**kwargs)

        return urllib.parse.urljoin(self.base_url, url), path.method


class Path:
    def __init__(self, path: dict) -> None:
        self.func = path[0]
        self.url = path[1]
        self.method = path[2]

    @property
    def name(self) -> str:
        return self.func

    def generate_url(self, **kwargs: typing.Any) -> str:
        parameters = {key: value for key, value in kwargs.items() if value}

        return self.url.format_map(defaultdict(str, **parameters))



import abc
import json
import logging
import typing
from collections import defaultdict
from urllib.parse import urlparse

import httpx
from httpx._client import UNSET, TimeoutTypes, UnsetType

from schema_registry.client import status, utils
from schema_registry.client.errors import ClientError
from schema_registry.client.paths import paths
from schema_registry.client.schema import BaseSchema, AvroSchema, JsonSchema, SchemaFactory, SubjectVersion
from schema_registry.client.urls import UrlManager

logger = logging.getLogger(__name__)


class BaseClient:
    """
    A client that talks to a Schema Registry over HTTP

    Args:
        url (str|dict) url: Url to schema registry or dictionary containing client configuration.
        ca_location (str): File or directory path to CA certificate(s)
            for verifying the Schema Registry key.
        cert_location (str): Path to public key used for authentication.
        key_location (str): Path to ./vate key used for authentication.
        key_password (str): Key password
        extra_headers (dict): Dictionary of HTTP headers to include when sending requests.
        timeout (httpx._client.TimeoutTypes): Optional. The timeout configuration to use when sending requests.
        pool_limits (httpx.Limits): Optional. The connection pool configuration to use when
            determining the maximum number of concurrently open HTTP connections.
    """

    def __init__(
        self,
        url: typing.Union[str, dict],
        ca_location: str = None,
        cert_location: str = None,
        key_location: str = None,
        key_password: str = None,
        extra_headers: dict = None,
        timeout: typing.Optional[httpx.Timeout] = None,
        pool_limits: typing.Optional[httpx.Limits] = None,
    ) -> None:

        if isinstance(url, str):
            conf = {
                utils.URL: url,
                utils.SSL_CA_LOCATION: ca_location,
                utils.SSL_CERTIFICATE_LOCATION: cert_location,
                utils.SSL_KEY_LOCATION: key_location,
                utils.SSL_KEY_PASSWORD: key_password,
            }
        else:
            conf = url

        self.conf = conf
        schema_server_url = conf.get(utils.URL, "")

        self.url_manager = UrlManager(schema_server_url, paths)  # type: ignore
        self.extra_headers = extra_headers
        self.timeout = timeout
        self.pool_limits = pool_limits

        self.session = self._create_session()

        # Cache Schemas: subj => { schema => id }
        self.subject_to_schema_ids: dict = defaultdict(dict)
        # Cache Schemas: id => avro_schema
        self.id_to_schema: dict = defaultdict(dict)
        # Cache Schemas: subj => { schema => version }
        self.subject_to_schema_versions: dict = defaultdict(dict)

    def __getstate__(self) -> typing.Dict:
        state = self.__dict__.copy()
        # Remove the unpicklable session.
        del state["session"]

        return state

    def __setstate__(self, state: typing.Dict) -> None:
        self.__dict__.update(state)
        self.session = self._create_session()

    def __eq__(self, obj: typing.Any) -> bool:
        return self.conf == obj.conf and self.extra_headers == obj.extra_headers

    @staticmethod
    def _configure_basic_auth(conf: typing.Dict[str, typing.Any]) -> typing.Tuple[str, str]:
        url = conf["url"]
        auth_provider = conf.pop("basic.auth.credentials.source", "URL").upper()

        if auth_provider not in utils.VALID_AUTH_PROVIDERS:
            raise ValueError(
                f"""
                schema.registry.basic.auth.credentials.source
                must be one of {utils.VALID_AUTH_PROVIDERS}
                """
            )

        if auth_provider == "USER_INFO":
            auth = tuple(conf.pop("basic.auth.user.info", "").split(":"))
        elif auth_provider == "SASL_INHERIT":
            if conf.pop("sasl.mechanism", "").upper() is ["GSSAPI"]:
                raise ValueError("SASL_INHERIT does not support SASL mechanisms GSSAPI")
            auth = (conf.pop("sasl.username", ""), conf.pop("sasl.password", ""))
        else:
            parsed_url = urlparse(url)
            auth = (parsed_url.username or "", parsed_url.password or "")

        # remove ignore after mypy fix https://github.com/python/mypy/issues/4805
        return auth  # type: ignore

    @staticmethod
    def _configure_client_tls(
        conf: dict,
    ) -> typing.Optional[typing.Union[str, typing.Tuple[str, str], typing.Tuple[str, str, str]]]:
        certificate = conf.get(utils.SSL_CERTIFICATE_LOCATION)

        if certificate:
            key_path = conf.get(utils.SSL_KEY_LOCATION)
            key_password = conf.get(utils.SSL_KEY_PASSWORD)

            if key_path:
                certificate = (certificate, key_path)

                if key_password:
                    certificate += (key_password,)

        return certificate

    def prepare_headers(self, body: dict = None, headers: dict = None) -> dict:
        _headers = {"Accept": utils.ACCEPT_HEADERS}

        if body:
            _headers["Content-Type"] = utils.HEADERS

        if headers:
            _headers.update(headers)

        return _headers

    @staticmethod
    def _add_to_cache(
        cache: dict, subject: str, schema: typing.Union[BaseSchema, str], value: typing.Union[str, int]
    ) -> None:
        sub_cache = cache[subject]
        sub_cache[schema] = value

    def _cache_schema(
        self,
        schema: typing.Union[BaseSchema, str],
        schema_id: int,
        subject: str = None,
        version: typing.Union[str, int] = None,
    ) -> None:
        if schema_id in self.id_to_schema:
            schema = self.id_to_schema[schema_id]
        else:
            self.id_to_schema[schema_id] = schema

        if subject:
            self._add_to_cache(self.subject_to_schema_ids, subject, schema, schema_id)
            if version:
                self._add_to_cache(self.subject_to_schema_versions, subject, schema, version)

    @abc.abstractmethod
    def _create_session(self):  # type: ignore
        ...


class SchemaRegistryClient(BaseClient):
    def _create_session(self) -> httpx.Client:
        """
        Create a httpx Client session
        Returns:
            httpx.Client
        """
        verify = self.conf.get(utils.SSL_CA_LOCATION, False)
        certificate = self._configure_client_tls(self.conf)
        auth = self._configure_basic_auth(self.conf)

        client_kwargs = {
            "cert": certificate,
            "verify": verify,  # type: ignore
            "auth": auth,
        }

        # If these values haven't been explicitly defined let httpx sort out
        # the default values.
        if self.extra_headers is not None:
            client_kwargs["headers"] = self.extra_headers  # type:ignore

        if self.timeout is not None:
            client_kwargs["timeout"] = self.timeout  # type:ignore

        if self.pool_limits is not None:
            client_kwargs["limits"] = self.pool_limits  # type:ignore

        return httpx.Client(**client_kwargs)  # type: ignore

    def request(
        self,
        url: str,
        method: str = "GET",
        body: dict = None,
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
    ) -> tuple:
        if method not in utils.VALID_METHODS:
            raise ClientError(f"Method {method} is invalid; valid methods include {utils.VALID_METHODS}")

        _headers = self.prepare_headers(body=body, headers=headers)

        with self.session as session:
            response = session.request(method, url, headers=_headers, json=body, timeout=timeout)

        try:
            return response.json(), response.status_code
        except ValueError:
            return response.content, response.status_code

    def register(
        self,
        subject: str,
        schema: typing.Union[BaseSchema, str],
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
        schema_type: str = utils.AVRO_SCHEMA_TYPE,
    ) -> int:
        """
        POST /subjects/(string: subject)/versions
        Register a schema with the registry under the given subject
        and receive a schema id.
        Schema can be avro or json, and can be presented as a parsed schema or a string.
        If schema is a string, the `schema_type` kwarg must be used to indicate
        what type of schema the string is ("AVRO" by default).
        If the schema is already parsed, the schema_type is inferred directly from the parsed schema.
        Multiple instances of the same schema will result in cache misses.

        Args:
            subject (str):
                subject name
            schema typing.Union[client.schema.BaseSchema, str]:
                Avro or JSON schema to be registered
            headers (dict):
                Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes):
                The timeout configuration to use when sending requests. Default UNSET
            schema_type typing.Union["AVRO", "JSON"]:
                The type of schema to parse if `schema` is a string. Default "AVRO"

        Returns:
            int: schema_id
        """
        schemas_to_id = self.subject_to_schema_ids[subject]

        if isinstance(schema, str):
            schema = SchemaFactory.create_schema(schema, schema_type)

        schema_id = schemas_to_id.get(schema)

        if schema_id is not None:
            return schema_id

        # Check if schema is already registered. This should normally work even if
        # the schema registry we're talking to is readonly, enabling a setup where
        # only CI/CD can do changes to Schema Registry, and production code has readonly
        # access

        response = self.check_version(subject, schema, headers=headers, timeout=timeout)

        if response is not None:
            return response.schema_id

        url, method = self.url_manager.url_for("register", subject=subject)
        body = {"schema": json.dumps(schema.raw_schema), "schemaType": schema.schema_type}

        result, code = self.request(url, method=method, body=body, headers=headers, timeout=timeout)

        msg = None
        if code in (status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN):
            msg = "Unauthorized access"
        elif code == status.HTTP_409_CONFLICT:
            msg = "Incompatible schema"
        elif code == status.HTTP_422_UNPROCESSABLE_ENTITY:
            msg = "Invalid schema"
        elif not status.is_success(code):
            msg = "Unable to register schema"

        if msg is not None:
            raise ClientError(message=msg, http_code=code, server_traceback=result)

        schema_id = result["id"]
        self._cache_schema(schema, schema_id, subject)

        return schema_id

    def get_subjects(self, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET) -> list:
        """
        GET /subjects/(string: subject)
        Get list of all registered subjects in your Schema Registry.

        Args:
            subject (str): subject name
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            list [str]: list of registered subjects.
        """
        url, method = self.url_manager.url_for("get_subjects")
        result, code = self.request(url, method=method, headers=headers, timeout=timeout)

        if status.is_success(code):
            return result

        raise ClientError("Unable to get subjects", http_code=code, server_traceback=result)

    def delete_subject(
        self, subject: str, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET
    ) -> list:
        """
        DELETE /subjects/(string: subject)
        Deletes the specified subject and its associated compatibility level if registered.
        It is recommended to use this API only when a topic needs to be
        recycled or in development environments.

        Args:
            subject (str): subject name
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            list (int): version of the schema deleted under this subject
        """
        url, method = self.url_manager.url_for("delete_subject", subject=subject)
        result, code = self.request(url, method=method, headers=headers, timeout=timeout)

        if status.is_success(code):
            return result
        elif code == status.HTTP_404_NOT_FOUND:
            return []

        raise ClientError("Unable to delete subject", http_code=code, server_traceback=result)

    def get_by_id(
        self, schema_id: int, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET
    ) -> typing.Optional[typing.Union[AvroSchema, JsonSchema]]:
        """
        GET /schemas/ids/{int: id}
        Retrieve a parsed avro schema by id or None if not found

        Args:
            schema_id (int): Schema Id
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            typing.Union[client.schema.AvroSchema, client.schema.JsonSchema]: Avro or JSON Record schema
        """
        if schema_id in self.id_to_schema:
            return self.id_to_schema[schema_id]

        url, method = self.url_manager.url_for("get_by_id", schema_id=schema_id)

        result, code = self.request(url, method=method, headers=headers, timeout=timeout)
        if code == status.HTTP_404_NOT_FOUND:
            logger.info(f"Schema {schema_id} not found: {code}")
            return None
        elif status.is_success(code):
            schema_str = result.get("schema")
            schema_type = result.get("schemaType", utils.AVRO_SCHEMA_TYPE)

            schema = SchemaFactory.create_schema(schema_str, schema_type)

            self._cache_schema(schema, schema_id)
            return schema

        raise ClientError(f"Received bad schema (id {schema_id})", http_code=code, server_traceback=result)

    def get_schema_subject_versions(
        self, schema_id: int, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET
    ) -> typing.Optional[typing.List[SubjectVersion]]:
        """
        GET /schemas/ids/{int: id}/versions
        Get the subject-version pairs identified by the input ID.

        Args:
            schema_id (int): Schema Id
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            typing.List[SubjectVersion]: List of Subject/Version pairs where Schema Id is registered
        """
        url, method = self.url_manager.url_for("get_schema_subject_versions", schema_id=schema_id)
        result, code = self.request(url, method=method, headers=headers, timeout=timeout)
        if code == status.HTTP_404_NOT_FOUND:
            logger.warning(f"Schema {schema_id} not found: {code}")
            return None
        elif status.is_success(code):
            ret = []
            for entry in result:
                ret.append(SubjectVersion(entry["subject"], entry["version"]))
            return ret

        raise ClientError(f"Received bad schema (id {schema_id})", http_code=code, server_traceback=result)

    def get_schema(
        self,
        subject: str,
        version: typing.Union[int, str] = "latest",
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
    ) -> typing.Optional[utils.SchemaVersion]:
        """
        GET /subjects/(string: subject)/versions/(versionId: version)
        Get a specific version of the schema registered under this subject

        Args:
            subject (str): subject name
            version (int, optional): version id. If is None, the latest schema is returned
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            SchemaVersion (nametupled): (subject, schema_id, schema, version)

            None: If server returns a not success response:
                404: Schema not found
                422: Unprocessable entity
                ~ (200 - 299): Not success
        """
        url, method = self.url_manager.url_for("get_schema", subject=subject, version=version)

        result, code = self.request(url, method=method, headers=headers, timeout=timeout)
        if code == status.HTTP_404_NOT_FOUND:
            logger.info(f"Schema version {version} under subjet {subject } not found: {code}")
            return None
        elif code == status.HTTP_422_UNPROCESSABLE_ENTITY:
            logger.info(f"Invalid version {version}: {code}")
            return None
        elif not status.is_success(code):
            logger.info(f"Not success version {version}: {code}")
            return None

        schema_id = result.get("id")
        if schema_id in self.id_to_schema:
            schema = self.id_to_schema[schema_id]
        else:
            schema = AvroSchema(result["schema"])

        version = result["version"]
        self._cache_schema(schema, schema_id, subject, version)

        return utils.SchemaVersion(subject, schema_id, schema, version)

    def get_versions(
        self, subject: str, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET
    ) -> list:
        """
        GET subjects/{subject}/versions
        Get a list of versions registered under the specified subject.

        Args:
            subject (str): subject name
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            list (str): version of the schema registered under this subject
        """
        url, method = self.url_manager.url_for("get_versions", subject=subject)

        result, code = self.request(url, method=method, headers=headers, timeout=timeout)
        if status.is_success(code):
            return result
        elif code == status.HTTP_404_NOT_FOUND:
            logger.info(f"Subject {subject} not found")
            return []

        raise ClientError(f"Unable to get the versions for subject {subject}", http_code=code, server_traceback=result)

    def delete_version(
        self,
        subject: str,
        version: typing.Union[int, str] = "latest",
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
    ) -> typing.Optional[int]:
        """
        DELETE /subjects/(string: subject)/versions/(versionId: version)
        Deletes a specific version of the schema registered under this subject.
        This only deletes the version and the schema ID remains intact making
        it still possible to decode data using the schema ID.
        This API is recommended to be used only in development environments or
        under extreme circumstances where-in, its required to delete a previously
        registered schema for compatibility purposes or re-register previously registered schema.

        Args:
            subject (str): subject name
            version (str): Version of the schema to be deleted.
                Valid values for versionId are between [1,2^31-1] or the string "latest".
                "latest" deletes the last registered schema under the specified subject.
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            int: version of the schema deleted
            None: If the subject or version does not exist.
        """
        url, method = self.url_manager.url_for("delete_version", subject=subject, version=version)

        result, code = self.request(url, method=method, headers=headers, timeout=timeout)

        if status.is_success(code):
            return result
        elif status.is_client_error(code):
            return None

        raise ClientError("Unable to delete the version", http_code=code, server_traceback=result)

    def check_version(
        self,
        subject: str,
        schema: typing.Union[BaseSchema, str],
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
        schema_type: str = utils.AVRO_SCHEMA_TYPE,
    ) -> typing.Optional[utils.SchemaVersion]:
        """
        POST /subjects/(string: subject)
        Check if a schema has already been registered under the specified subject.
        If so, this returns the schema string along with its globally unique identifier,
        its version under this subject and the subject name.

        Args:
            subject (str):
                subject name
            schema typing.Union[client.schema.BaseSchema, str]:
                Avro or JSON schema
            headers (dict):
                Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes):
                The timeout configuration to use when sending requests. Default UNSET
            schema_type typing.Union["AVRO", "JSON"]:
                The type of schema to parse if `schema` is a string. Default "AVRO"

        Returns:
            SchemaVersion (nametupled): (subject, schema_id, schema, version)
            None: If schema not found.
        """
        schemas_to_version = self.subject_to_schema_versions[subject]

        if isinstance(schema, str):
            schema = SchemaFactory.create_schema(schema, schema_type)

        version = schemas_to_version.get(schema)

        schemas_to_id = self.subject_to_schema_ids[subject]
        schema_id = schemas_to_id.get(schema)

        if all((version, schema_id)):
            return utils.SchemaVersion(subject, schema_id, version, schema)

        url, method = self.url_manager.url_for("check_version", subject=subject)
        body = {"schema": json.dumps(schema.raw_schema), "schemaType": schema.schema_type}

        result, code = self.request(url, method=method, body=body, headers=headers, timeout=timeout)
        if code == status.HTTP_404_NOT_FOUND:
            logger.info(f"Schema {schema.name} under subject {subject} not found: {code}")
            return None
        elif status.is_success(code):
            schema_id = result.get("id")
            version = result.get("version")
            self._cache_schema(schema, schema_id, subject, version)

            return utils.SchemaVersion(subject, schema_id, version, result.get("schema"))

        raise ClientError("Unable to get version of a schema", http_code=code, server_traceback=result)

    def test_compatibility(
        self,
        subject: str,
        schema: typing.Union[AvroSchema, JsonSchema, str],
        version: typing.Union[int, str] = "latest",
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
        schema_type: str = utils.AVRO_SCHEMA_TYPE,
    ) -> bool:
        """
        POST /compatibility/subjects/(string: subject)/versions/(versionId: version)
        Test the compatibility of a candidate parsed schema for a given subject.
        By default the latest version is checked against.

        Args:
            subject (str):
                subject name
            schema typing.Union[client.schema.AvroSchema, client.schema.JsonSchema, str]:
                Avro or JSON schema
            headers (dict):
                Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes):
                The timeout configuration to use when sending requests. Default UNSET
            schema_type typing.Union["AVRO", "JSON"]:
                The type of schema to parse if `schema` is a string. Default "AVRO"

        Returns:
            bool: True if schema given compatible, False otherwise
        """
        url, method = self.url_manager.url_for("test_compatibility", subject=subject, version=version)

        if isinstance(schema, str):
            schema = SchemaFactory.create_schema(schema, schema_type)

        body = {"schema": json.dumps(schema.raw_schema), "schemaType": schema.schema_type}
        result, code = self.request(url, method=method, body=body, headers=headers, timeout=timeout)

        if code == status.HTTP_404_NOT_FOUND:
            logger.info(f"Subject or version not found: {code}")
            return False
        elif code == status.HTTP_422_UNPROCESSABLE_ENTITY:
            logger.info(f"Unprocessable entity. Invalid subject or schema: {code}")
        elif status.is_success(code):
            return result.get("is_compatible")

        raise ClientError("Unable to check the compatibility", http_code=code, server_traceback=result)

    def update_compatibility(
        self,
        level: str,
        subject: str = None,
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
    ) -> bool:
        """
        PUT /config/(string: subject)
        Update the compatibility level.
        If subject is None, the compatibility level is global.

        Args:
            level (str): one of BACKWARD, BACKWARD_TRANSITIVE, FORWARD, FORWARD_TRANSITIVE,
                FULL, FULL_TRANSITIVE, NONE
            subject (str): Option subject
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            bool: True if compatibility was updated

        Raises:
            ClientError: if the request was unsuccessful or an invalid
        """
        if level not in utils.VALID_LEVELS:
            raise ClientError(f"Invalid level specified: {level}")

        url, method = self.url_manager.url_for("update_compatibility", subject=subject)
        body = {"compatibility": level}

        result, code = self.request(url, method=method, body=body, headers=headers, timeout=timeout)

        if status.is_success(code):
            return True

        raise ClientError(f"Unable to update level: {level}.", http_code=code, server_traceback=result)

    def get_compatibility(
        self, subject: str = None, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET
    ) -> str:
        """
        Get the current compatibility level for a subject.

        Args:
            subject (str): subject name
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            str: one of BACKWARD, BACKWARD_TRANSITIVE, FORWARD, FORWARD_TRANSITIVE,
                FULL, FULL_TRANSITIVE, NONE

        Raises:
            ClientError: if the request was unsuccessful or an invalid
            compatibility level was returned
        """
        url, method = self.url_manager.url_for("get_compatibility", subject=subject)
        result, code = self.request(url, method=method, headers=headers, timeout=timeout)

        if status.is_success(code):
            compatibility = result.get("compatibilityLevel")
            if compatibility not in utils.VALID_LEVELS:
                if compatibility is None:
                    error_msg_suffix = "No compatibility was returned"
                else:
                    error_msg_suffix = str(compatibility)
                raise ClientError(
                    f"Invalid compatibility level received: {error_msg_suffix}", http_code=code, server_traceback=result
                )

            return compatibility

        raise ClientError(
            f"Unable to fetch compatibility level. Error code: {code}", http_code=code, server_traceback=result
        )


class AsyncSchemaRegistryClient(BaseClient):
    def _create_session(self) -> httpx.AsyncClient:
        """
        Create a httpx Client session
        Returns:
            httpx.AsyncClient
        """
        verify = self.conf.get(utils.SSL_CA_LOCATION, False)
        certificate = self._configure_client_tls(self.conf)
        auth = self._configure_basic_auth(self.conf)

        client_kwargs = {
            "cert": certificate,
            "verify": verify,  # type: ignore
            "auth": auth,
        }

        # If these values haven't been explicitly defined let httpx sort out
        # the default values.
        if self.extra_headers is not None:
            client_kwargs["headers"] = self.extra_headers  # type:ignore

        if self.timeout is not None:
            client_kwargs["timeout"] = self.timeout  # type:ignore

        if self.pool_limits is not None:
            client_kwargs["limits"] = self.pool_limits  # type:ignore

        return httpx.AsyncClient(**client_kwargs)  # type: ignore

    async def request(
        self,
        url: str,
        method: str = "GET",
        body: dict = None,
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
    ) -> tuple:
        if method not in utils.VALID_METHODS:
            raise ClientError(f"Method {method} is invalid; valid methods include {utils.VALID_METHODS}")

        _headers = self.prepare_headers(body=body, headers=headers)

        async with self.session as session:
            response = await session.request(method, url, headers=_headers, json=body, timeout=timeout)

        try:
            return response.json(), response.status_code
        except ValueError:
            return response.content, response.status_code

    async def register(
        self,
        subject: str,
        schema: typing.Union[BaseSchema, str],
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
        schema_type: str = utils.AVRO_SCHEMA_TYPE,
    ) -> int:
        """
        POST /subjects/(string: subject)/versions
        Register a schema with the registry under the given subject
        and receive a schema id.
        Schema can be avro or json, and can be presented as a parsed schema or a string.
        If schema is a string, the `schema_type` kwarg must be used to indicate
        what type of schema the string is ("AVRO" by default).
        If the schema is already parsed, the schema_type is inferred directly from the parsed schema.
        Multiple instances of the same schema will result in cache misses.


        Args:
            subject (str):
                subject name
            schema typing.Union[client.schema.AvroSchema, client.schema.JsonSchema, str]:
                Avro or JSON schema
            headers (dict):
                Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes):
                The timeout configuration to use when sending requests. Default UNSET
            schema_type typing.Union["AVRO", "JSON"]:
                The type of schema to parse if `schema` is a string. Default "AVRO"

        Returns:
            int: schema_id
        """
        schemas_to_id = self.subject_to_schema_ids[subject]

        if isinstance(schema, str):
            schema = SchemaFactory.create_schema(schema, schema_type)

        schema_id = schemas_to_id.get(schema)

        if schema_id is not None:
            return schema_id

        # Check if schema is already registered. This should normally work even if
        # the schema registry we're talking to is readonly, enabling a setup where
        # only CI/CD can do changes to Schema Registry, and production code has readonly
        # access

        response = await self.check_version(subject, schema, headers=headers, timeout=timeout)

        if response is not None:
            return response.schema_id

        url, method = self.url_manager.url_for("register", subject=subject)
        body = {"schema": json.dumps(schema.raw_schema), "schemaType": schema.schema_type}

        result, code = await self.request(url, method=method, body=body, headers=headers, timeout=timeout)

        msg = None
        if code in (status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN):
            msg = "Unauthorized access"
        elif code == status.HTTP_409_CONFLICT:
            msg = "Incompatible Avro schema"
        elif code == status.HTTP_422_UNPROCESSABLE_ENTITY:
            msg = "Invalid Avro schema"
        elif not status.is_success(code):
            msg = "Unable to register schema"

        if msg is not None:
            raise ClientError(message=msg, http_code=code, server_traceback=result)

        schema_id = result["id"]
        self._cache_schema(schema, schema_id, subject)

        return schema_id

    async def get_subjects(self, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET) -> list:
        """
        GET /subjects/(string: subject)
        Get list of all registered subjects in your Schema Registry.

        Args:
            subject (str): subject name
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            list [str]: list of registered subjects.
        """
        url, method = self.url_manager.url_for("get_subjects")
        result, code = await self.request(url, method=method, headers=headers, timeout=timeout)

        if status.is_success(code):
            return result

        raise ClientError("Unable to get subjects", http_code=code, server_traceback=result)

    async def delete_subject(
        self, subject: str, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET
    ) -> list:
        """
        DELETE /subjects/(string: subject)
        Deletes the specified subject and its associated compatibility level if registered.
        It is recommended to use this API only when a topic needs to be
        recycled or in development environments.

        Args:
            subject (str): subject name
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            list (int): version of the schema deleted under this subject
        """
        url, method = self.url_manager.url_for("delete_subject", subject=subject)
        result, code = await self.request(url, method=method, headers=headers, timeout=timeout)

        if status.is_success(code):
            return result
        elif code == status.HTTP_404_NOT_FOUND:
            return []

        raise ClientError("Unable to delete subject", http_code=code, server_traceback=result)

    async def get_by_id(
        self, schema_id: int, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET
    ) -> typing.Optional[typing.Union[AvroSchema, JsonSchema]]:
        """
        GET /schemas/ids/{int: id}
        Retrieve a parsed avro schema by id or None if not found

        Args:
            schema_id (int): Schema Id
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            typing.Union[client.schema.AvroSchema, client.schema.JsonSchema]: Avro or JSON Record schema
        """
        if schema_id in self.id_to_schema:
            return self.id_to_schema[schema_id]

        url, method = self.url_manager.url_for("get_by_id", schema_id=schema_id)

        result, code = await self.request(url, method=method, headers=headers, timeout=timeout)
        if code == status.HTTP_404_NOT_FOUND:
            logger.info(f"Schema {schema_id} not found: {code}")
            return None
        elif status.is_success(code):
            schema_str = result.get("schema")
            result = AvroSchema(schema_str)

            self._cache_schema(result, schema_id)
            return result

        raise ClientError(f"Received bad schema (id {schema_id})", http_code=code, server_traceback=result)

    async def get_schema(
        self,
        subject: str,
        version: typing.Union[int, str] = "latest",
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
    ) -> typing.Optional[utils.SchemaVersion]:
        """
        GET /subjects/(string: subject)/versions/(versionId: version)
        Get a specific version of the schema registered under this subject

        Args:
            subject (str): subject name
            version (int, optional): version id. If is None, the latest schema is returned
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            SchemaVersion (nametupled): (subject, schema_id, schema, version)

            None: If server returns a not success response:
                404: Schema not found
                422: Unprocessable entity
                ~ (200 - 299): Not success
        """
        url, method = self.url_manager.url_for("get_schema", subject=subject, version=version)

        result, code = await self.request(url, method=method, headers=headers, timeout=timeout)

        if code == status.HTTP_404_NOT_FOUND:
            logger.info(f"Schema version {version} under subjet {subject } not found: {code}")
            return None
        elif code == status.HTTP_422_UNPROCESSABLE_ENTITY:
            logger.info(f"Invalid version {version}: {code}")
            return None
        elif not status.is_success(code):
            logger.info(f"Not success version {version}: {code}")
            return None

        schema_id = result.get("id")
        if schema_id in self.id_to_schema:
            schema = self.id_to_schema[schema_id]
        else:
            schema = AvroSchema(result["schema"])

        version = result["version"]
        self._cache_schema(schema, schema_id, subject, version)

        return utils.SchemaVersion(subject, schema_id, schema, version)

    async def get_schema_subject_versions(
        self, schema_id: int, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET
    ) -> typing.Optional[typing.List[SubjectVersion]]:
        """
        GET /schemas/ids/{int: id}/versions
        Get the subject-version pairs identified by the input ID.

        Args:
            schema_id (int): Schema Id
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            typing.List[SubjectVersion]: List of Subject/Version pairs where Schema Id is registered
        """
        url, method = self.url_manager.url_for("get_schema_subject_versions", schema_id=schema_id)
        result, code = await self.request(url, method=method, headers=headers, timeout=timeout)

        if code == status.HTTP_404_NOT_FOUND:
            logger.warning(f"Schema {schema_id} not found: {code}")
            return None
        elif status.is_success(code):
            ret = []
            for entry in result:
                ret.append(SubjectVersion(entry["subject"], entry["version"]))
            return ret

        raise ClientError(f"Received bad schema (id {schema_id})", http_code=code, server_traceback=result)

    async def get_versions(
        self, subject: str, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET
    ) -> list:
        """
        GET subjects/{subject}/versions
        Get a list of versions registered under the specified subject.

        Args:
            subject (str): subject name
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            list (str): version of the schema registered under this subject
        """
        url, method = self.url_manager.url_for("get_versions", subject=subject)

        result, code = await self.request(url, method=method, headers=headers, timeout=timeout)
        if status.is_success(code):
            return result
        elif code == status.HTTP_404_NOT_FOUND:
            logger.info(f"Subject {subject} not found")
            return []

        raise ClientError(f"Unable to get the versions for subject {subject}", http_code=code, server_traceback=result)

    async def delete_version(
        self,
        subject: str,
        version: typing.Union[int, str] = "latest",
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
    ) -> typing.Optional[int]:
        """
        DELETE /subjects/(string: subject)/versions/(versionId: version)
        Deletes a specific version of the schema registered under this subject.
        This only deletes the version and the schema ID remains intact making
        it still possible to decode data using the schema ID.
        This API is recommended to be used only in development environments or
        under extreme circumstances where-in, its required to delete a previously
        registered schema for compatibility purposes or re-register previously registered schema.

        Args:
            subject (str): subject name
            version (str): Version of the schema to be deleted.
                Valid values for versionId are between [1,2^31-1] or the string "latest".
                "latest" deletes the last registered schema under the specified subject.
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            int: version of the schema deleted
            None: If the subject or version does not exist.
        """
        url, method = self.url_manager.url_for("delete_version", subject=subject, version=version)

        result, code = await self.request(url, method=method, headers=headers, timeout=timeout)

        if status.is_success(code):
            return result
        elif status.is_client_error(code):
            return None

        raise ClientError("Unable to delete the version", http_code=code, server_traceback=result)

    async def check_version(
        self,
        subject: str,
        schema: typing.Union[BaseSchema, str],
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
        schema_type: str = utils.AVRO_SCHEMA_TYPE,
    ) -> typing.Optional[utils.SchemaVersion]:
        """
        POST /subjects/(string: subject)
        Check if a schema has already been registered under the specified subject.
        If so, this returns the schema string along with its globally unique identifier,
        its version under this subject and the subject name.

        Args:
            subject (str):
                subject name
            schema typing.Union[client.schema.AvroSchema, client.schema.JsonSchema, str]:
                Avro or JSON schema
            headers (dict):
                Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes):
                The timeout configuration to use when sending requests. Default UNSET
            schema_type typing.Union["AVRO", "JSON"]:
                The type of schema to parse if `schema` is a string. Default "AVRO"

        Returns:
            SchemaVersion (nametupled): (subject, schema_id, schema, version)
            None: If schema not found.
        """
        schemas_to_version = self.subject_to_schema_versions[subject]

        if isinstance(schema, str):
            schema = SchemaFactory.create_schema(schema, schema_type)

        version = schemas_to_version.get(schema)

        schemas_to_id = self.subject_to_schema_ids[subject]
        schema_id = schemas_to_id.get(schema)

        if all((version, schema_id)):
            return utils.SchemaVersion(subject, schema_id, version, schema)

        url, method = self.url_manager.url_for("check_version", subject=subject)
        body = {"schema": json.dumps(schema.raw_schema), "schemaType": schema.schema_type}

        result, code = await self.request(url, method=method, body=body, headers=headers, timeout=timeout)
        if code == status.HTTP_404_NOT_FOUND:
            logger.info(f"Schema {schema.name} under subject {subject} not found: {code}")
            return None
        elif status.is_success(code):
            schema_id = result.get("id")
            version = result.get("version")
            self._cache_schema(schema, schema_id, subject, version)

            return utils.SchemaVersion(subject, schema_id, version, result.get("schema"))

        raise ClientError("Unable to get version of a schema", http_code=code, server_traceback=result)

    async def test_compatibility(
        self,
        subject: str,
        schema: typing.Union[AvroSchema, JsonSchema, str],
        version: typing.Union[int, str] = "latest",
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
        schema_type: str = utils.AVRO_SCHEMA_TYPE,
    ) -> bool:
        """
        POST /compatibility/subjects/(string: subject)/versions/(versionId: version)
        Test the compatibility of a candidate parsed schema for a given subject.
        By default the latest version is checked against.

        Args:
            subject (str):
                subject name
            schema typing.Union[client.schema.AvroSchema, client.schema.JsonSchema, str]:
                Avro or JSON schema
            headers (dict):
                Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes):
                The timeout configuration to use when sending requests. Default UNSET
            schema_type typing.Union["AVRO", "JSON"]:
                The type of schema to parse if `schema` is a string. Default "AVRO"

        Returns:
            bool: True if schema given compatible, False otherwise
        """
        url, method = self.url_manager.url_for("test_compatibility", subject=subject, version=version)

        if isinstance(schema, str):
            schema = SchemaFactory.create_schema(schema, schema_type)

        body = {"schema": json.dumps(schema.raw_schema), "schemaType": schema.schema_type}
        result, code = await self.request(url, method=method, body=body, headers=headers, timeout=timeout)

        if code == status.HTTP_404_NOT_FOUND:
            logger.info(f"Subject or version not found: {code}")
            return False
        elif code == status.HTTP_422_UNPROCESSABLE_ENTITY:
            logger.info(f"Unprocessable entity. Invalid subject or schema: {code}")
            return False
        elif status.is_success(code):
            return result.get("is_compatible")

        raise ClientError("Unable to check the compatibility", http_code=code, server_traceback=result)

    async def update_compatibility(
        self,
        level: str,
        subject: str = None,
        headers: dict = None,
        timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET,
    ) -> bool:
        """
        PUT /config/(string: subject)
        Update the compatibility level.
        If subject is None, the compatibility level is global.

        Args:
            level (str): one of BACKWARD, BACKWARD_TRANSITIVE, FORWARD, FORWARD_TRANSITIVE,
                FULL, FULL_TRANSITIVE, NONE
            subject (str): Option subject
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            bool: True if compatibility was updated

        Raises:
            ClientError: if the request was unsuccessful or an invalid
        """
        if level not in utils.VALID_LEVELS:
            raise ClientError(f"Invalid level specified: {level}")

        url, method = self.url_manager.url_for("update_compatibility", subject=subject)
        body = {"compatibility": level}

        result, code = await self.request(url, method=method, body=body, headers=headers, timeout=timeout)

        if status.is_success(code):
            return True

        raise ClientError(f"Unable to update level: {level}.", http_code=code, server_traceback=result)

    async def get_compatibility(
        self, subject: str = None, headers: dict = None, timeout: typing.Union[TimeoutTypes, UnsetType] = UNSET
    ) -> str:
        """
        Get the current compatibility level for a subject.

        Args:
            subject (str): subject name
            headers (dict): Extra headers to add on the requests
            timeout (httpx._client.TimeoutTypes): The timeout configuration to use when sending requests. Default UNSET

        Returns:
            str: one of BACKWARD, BACKWARD_TRANSITIVE, FORWARD, FORWARD_TRANSITIVE,
                FULL, FULL_TRANSITIVE, NONE

        Raises:
            ClientError: if the request was unsuccessful or an invalid
            compatibility level was returned
        """
        url, method = self.url_manager.url_for("get_compatibility", subject=subject)
        result, code = await self.request(url, method=method, headers=headers, timeout=timeout)

        if status.is_success(code):
            compatibility = result.get("compatibilityLevel")
            if compatibility not in utils.VALID_LEVELS:
                if compatibility is None:
                    error_msg_suffix = "No compatibility was returned"
                else:
                    error_msg_suffix = str(compatibility)
                raise ClientError(
                    f"Invalid compatibility level received: {error_msg_suffix}", http_code=code, server_traceback=result
                )

            return compatibility

        raise ClientError(
            f"Unable to fetch compatibility level. Error code: {code}", http_code=code, server_traceback=result
        )



import typing

import faust
import pydantic
from dataclasses_avroschema import AvroModel

from schema_registry.client import schema
from schema_registry.serializers import faust as serializer, AvroMessageSerializer, JsonMessageSerializer
from tests import data_gen


def test_create_avro_faust_serializer(client, avro_country_schema):
    schema_subject = "test-avro-country"
    faust_serializer = serializer.FaustSerializer(client, schema_subject, avro_country_schema)

    assert type(faust_serializer.message_serializer) == AvroMessageSerializer
    assert faust_serializer.schema_subject == schema_subject
    assert faust_serializer.schema == avro_country_schema
    assert faust_serializer.message_serializer.schemaregistry_client == client


def test_avro_dumps_load_message(client, avro_country_schema):
    faust_serializer = serializer.FaustSerializer(client, "test-avro-country", avro_country_schema)

    record = {"country": "Argentina"}
    message_encoded = faust_serializer._dumps(record)

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    message_decoded = faust_serializer._loads(message_encoded)
    assert message_decoded == record


def test_avro_nested_schema(client):
    nested_schema = schema.AvroSchema(data_gen.AVRO_NESTED_SCHENA)
    faust_serializer = serializer.FaustSerializer(client, "test-avro-nested-schema", nested_schema)

    record = data_gen.create_nested_schema()
    message_encoded = faust_serializer._dumps(record)

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    message_decoded = faust_serializer._loads(message_encoded)
    assert message_decoded == record


def test_avro_dumps_load_with_register_codec(client, avro_country_schema):
    payload = {"country": "Argentina"}
    country_serializer = serializer.FaustSerializer(client, "test-avro-country", avro_country_schema)

    faust.serializers.codecs.register("country_avro_serializer", country_serializer)

    class CountryRecord(faust.Record, serializer="country_avro_serializer"):
        country: str

    country_record = CountryRecord(**payload)
    message_encoded = country_record.dumps()

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    message_decoded = CountryRecord.loads(message_encoded)

    assert message_decoded == country_record


def test_avro_nested_schema_with_register_codec(client):
    nested_schema = schema.AvroSchema(data_gen.AVRO_NESTED_SCHENA)
    order_schema = schema.AvroSchema(data_gen.AVRO_ORDER_SCHENA)

    customer_serializer = serializer.FaustSerializer(client, "test-avro-nested-schema", nested_schema)
    order_serializer = serializer.FaustSerializer(client, "test-avro-order-schema", order_schema)

    faust.serializers.codecs.register("customer_avro_serializer", customer_serializer)
    faust.serializers.codecs.register("order_avro_serializer", order_serializer)

    class Order(faust.Record, serializer="order_avro_serializer"):
        uid: int

    class Customer(faust.Record, serializer="customer_avro_serializer"):
        name: str
        uid: int
        order: Order

    payload = data_gen.create_nested_schema()

    customer = Customer(**payload)

    message_encoded = customer.dumps()

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    message_decoded = Customer.loads(message_encoded)
    assert message_decoded == customer


def test_avro_dumps_load_message_dataclasses_avro_schema(client):
    class AdvanceUserModel(faust.Record, AvroModel):
        first_name: str
        last_name: str
        age: int

    faust_serializer = serializer.FaustSerializer(client, "test-dataclasses-avroschema", AdvanceUserModel.avro_schema())

    record = {
        "first_name": "Juan",
        "last_name": "Perez",
        "age": 20,
    }

    message_encoded = faust_serializer._dumps(record)

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    message_decoded = faust_serializer._loads(message_encoded)
    assert message_decoded == record


def test_avro_dumps_load_message_union_avro_schema(client):
    class FirstMemberRecord(faust.Record, AvroModel):
        name: str = ""

    class SecondMemberRecord(faust.Record, AvroModel):
        name: str = ""

    class UnionFieldAvroModel(faust.Record, AvroModel):
        a_name: typing.Union[FirstMemberRecord, SecondMemberRecord, None]

    avro_name = "test-union-field-avroschema"
    avro_schema = UnionFieldAvroModel.avro_schema()

    faust_serializer = serializer.FaustSerializer(client, avro_name, avro_schema, return_record_name=True)

    record = {"a_name": ("FirstMemberRecord", {"name": "jj"})}

    message_encoded = faust_serializer._dumps(record)

    assert message_encoded

    message_decoded = faust_serializer._loads(message_encoded)
    assert message_decoded == record


def test_create_json_faust_serializer(client, json_country_schema):
    schema_subject = "test-json-country"
    faust_serializer = serializer.FaustJsonSerializer(client, schema_subject, json_country_schema)

    assert type(faust_serializer.message_serializer) == JsonMessageSerializer
    assert faust_serializer.schema_subject == schema_subject
    assert faust_serializer.schema == json_country_schema
    assert faust_serializer.message_serializer.schemaregistry_client == client


def test_json_dumps_load_message(client, json_country_schema):
    faust_serializer = serializer.FaustJsonSerializer(client, "test-json-country", json_country_schema)

    record = {"country": "Argentina"}
    message_encoded = faust_serializer._dumps(record)

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    message_decoded = faust_serializer._loads(message_encoded)
    assert message_decoded == record


def test_json_nested_schema(client):
    nested_schema = schema.JsonSchema(data_gen.JSON_NESTED_SCHEMA)
    faust_serializer = serializer.FaustJsonSerializer(client, "test-json-nested-schema", nested_schema)

    record = data_gen.create_nested_schema()
    message_encoded = faust_serializer._dumps(record)

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    message_decoded = faust_serializer._loads(message_encoded)
    assert message_decoded == record


def test_json_dumps_load_with_register_codec(client, json_country_schema):
    payload = {"country": "Argentina"}
    country_serializer = serializer.FaustJsonSerializer(client, "test-json-country", json_country_schema)

    faust.serializers.codecs.register("country_json_serializer", country_serializer)

    class CountryRecord(faust.Record, serializer="country_json_serializer"):
        country: str

    country_record = CountryRecord(**payload)
    message_encoded = country_record.dumps()

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    message_decoded = CountryRecord.loads(message_encoded)

    assert message_decoded == country_record


def test_json_nested_schema_with_register_codec(client):
    nested_schema = schema.JsonSchema(data_gen.JSON_NESTED_SCHEMA)
    order_schema = schema.JsonSchema(data_gen.JSON_ORDER_SCHEMA)

    customer_serializer = serializer.FaustJsonSerializer(client, "test-json-nested-schema", nested_schema)
    order_serializer = serializer.FaustJsonSerializer(client, "test-json-order-schema", order_schema)

    faust.serializers.codecs.register("customer_json_serializer", customer_serializer)
    faust.serializers.codecs.register("order_json_serializer", order_serializer)

    class Order(faust.Record, serializer="order_json_serializer"):
        uid: int

    class Customer(faust.Record, serializer="customer_json_serializer"):
        name: str
        uid: int
        order: Order

    payload = data_gen.create_nested_schema()

    customer = Customer(**payload)

    message_encoded = customer.dumps()

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    message_decoded = Customer.loads(message_encoded)
    assert message_decoded == customer


def test_json_dumps_load_message_dataclasses_json_schema(client):
    class AdvanceUserModel(faust.Record, pydantic.BaseModel):
        first_name: str
        last_name: str
        age: int

    faust_serializer = serializer.FaustJsonSerializer(
        client, "test-dataclasses-jsonschema", AdvanceUserModel.schema_json()
    )

    record = {
        "first_name": "Juan",
        "last_name": "Perez",
        "age": 20,
    }

    message_encoded = faust_serializer._dumps(record)

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    message_decoded = faust_serializer._loads(message_encoded)
    assert message_decoded == record


def test_json_dumps_load_message_union_json_schema(client):
    class FirstMemberRecord(pydantic.BaseModel):
        name: str = ""

    class SecondMemberRecord(pydantic.BaseModel):
        name: str = ""

    class UnionFieldJsonModel(pydantic.BaseModel):
        a_name: typing.Union[FirstMemberRecord, SecondMemberRecord, None]

    json_name = "test-union-field-jsonschema"
    json_schema = UnionFieldJsonModel.schema_json()

    faust_serializer = serializer.FaustJsonSerializer(client, json_name, json_schema, return_record_name=True)

    record = {"a_name": {"name": "jj"}}

    message_encoded = faust_serializer._dumps(record)

    assert message_encoded

    message_decoded = faust_serializer._loads(message_encoded)
    assert message_decoded == record



import fastavro
import jsonschema
import pytest

from schema_registry.client import schema
from tests import data_gen


def test_avro_schema_from_string():
    parsed = schema.AvroSchema(data_gen.AVRO_BASIC_SCHEMA)

    assert isinstance(parsed, schema.AvroSchema)


@pytest.mark.asyncio
async def test_avro_schema_from_file():
    parsed = await schema.AvroSchema.async_load(data_gen.get_schema_path("adv_schema.avsc"))
    assert isinstance(parsed, schema.AvroSchema)


@pytest.mark.asyncio
async def test_avro_schema_load_parse_error():
    with pytest.raises(fastavro.schema.UnknownType):
        await schema.AvroSchema.async_load(data_gen.get_schema_path("invalid_schema.avsc"))


def test_json_schema_from_string():
    parsed = schema.JsonSchema(data_gen.JSON_BASIC_SCHEMA)

    assert isinstance(parsed, schema.JsonSchema)


@pytest.mark.asyncio
async def test_json_schema_from_file():
    parsed = await schema.JsonSchema.async_load(data_gen.get_schema_path("adv_schema.json"))
    assert isinstance(parsed, schema.JsonSchema)


@pytest.mark.asyncio
async def test_json_schema_load_parse_error():
    with pytest.raises(jsonschema.exceptions.SchemaError):
        await schema.JsonSchema.async_load(data_gen.get_schema_path("invalid_schema.json"))



class SerializerError(Exception):
    """Generic error from serializer package"""

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(error={self.message})"

    def __str__(self) -> str:
        return self.message



import fastavro
import jsonschema
import pytest

from schema_registry.client import schema, utils
from tests import data_gen


def test_avro_schema_from_string():
    parsed = schema.AvroSchema(data_gen.AVRO_BASIC_SCHEMA)
    assert isinstance(parsed, schema.AvroSchema)


def test_avro_schema_from_file():
    parsed = schema.AvroSchema.load(data_gen.get_schema_path("adv_schema.avsc"))
    assert isinstance(parsed, schema.AvroSchema)


def test_avro_schema_load_parse_error():
    with pytest.raises(fastavro.schema.UnknownType):
        schema.AvroSchema.load(data_gen.get_schema_path("invalid_schema.avsc"))


def test_avro_schema_type_property():
    parsed = schema.AvroSchema(data_gen.AVRO_BASIC_SCHEMA)
    assert parsed.schema_type == utils.AVRO_SCHEMA_TYPE


def test_expanded_schema(client):
    advance_schema = schema.AvroSchema(data_gen.AVRO_ADVANCED_SCHEMA)
    expanded = {
        "type": "record",
        "doc": "advanced schema for tests",
        "name": "python.test.advanced.advanced",
        "fields": [
            {"name": "number", "doc": "age", "type": ["long", "null"]},
            {"name": "name", "doc": "a name", "type": ["string"]},
            {
                "doc": "friends",
                "name": "friends",
                "type": {
                    "type": "map",
                    "values": {
                        "type": "record",
                        "name": "python.test.advanced.basicPerson",
                        "fields": [
                            {"doc": "friend age", "name": "number", "type": ["long", "null"]},
                            {"doc": "friend name", "name": "name", "type": ["string"]},
                        ],
                    },
                },
            },
            {
                "name": "family",
                "doc": "family",
                "type": {
                    "type": "map",
                    "values": {
                        "type": "record",
                        "name": "python.test.advanced.basicPerson",
                        "fields": [
                            {"doc": "friend age", "name": "number", "type": ["long", "null"]},
                            {"doc": "friend name", "name": "name", "type": ["string"]},
                        ],
                    },
                },
            },
        ],
    }

    assert advance_schema.expanded_schema == expanded


def test_flat_schema(client):
    advance_schema = schema.AvroSchema(data_gen.AVRO_ADVANCED_SCHEMA)
    subject = "test-avro-advance-schema"
    client.register(subject, advance_schema)

    schema_version = client.get_schema(subject)
    parsed_schema = schema_version.schema
    parsed_schema.schema.pop("__fastavro_parsed")
    parsed_schema.schema.pop("__named_schemas")

    assert schema_version.schema.flat_schema == parsed_schema.schema


def test_json_schema_from_string():
    parsed = schema.JsonSchema(data_gen.JSON_BASIC_SCHEMA)
    assert isinstance(parsed, schema.JsonSchema)


def test_json_schema_from_file():
    parsed = schema.JsonSchema.load(data_gen.get_schema_path("adv_schema.json"))
    assert isinstance(parsed, schema.JsonSchema)


def test_json_schema_load_parse_error():
    with pytest.raises(jsonschema.exceptions.SchemaError):
        schema.JsonSchema.load(data_gen.get_schema_path("invalid_schema.json"))


def test_json_schema_type_property():
    parsed = schema.JsonSchema(data_gen.JSON_BASIC_SCHEMA)
    assert parsed.schema_type == "JSON"



import pytest

from schema_registry.client import schema
from tests import data_gen


@pytest.mark.asyncio
async def test_avro_delete_subject(async_client, avro_user_schema_v3):
    subject = "avro-subject-to-delete"
    versions = [schema.AvroSchema(data_gen.AVRO_USER_V1), schema.AvroSchema(data_gen.AVRO_USER_V2)]

    for version in versions:
        await async_client.register(subject, version)

    assert len(await async_client.delete_subject(subject)) == len(versions)


@pytest.mark.asyncio
async def test_json_delete_subject(async_client, json_user_schema_v3):
    subject = "json-subject-to-delete"
    versions = [schema.JsonSchema(data_gen.JSON_USER_V2), json_user_schema_v3]

    for version in versions:
        await async_client.register(subject, version)

    assert len(await async_client.delete_subject(subject)) == len(versions)


@pytest.mark.asyncio
async def test_delete_subject_does_not_exist(async_client):
    assert not await async_client.delete_subject("a-random-subject")



import httpx
import pytest

from schema_registry.client import errors, schema, utils
from tests import data_gen


def test_avro_compatibility(client, avro_user_schema_v3):
    """
    Test the compatibility of a new User Schema against the User schema version 2.
    """
    subject = "test-avro-user-schema"
    version_2 = schema.AvroSchema(data_gen.AVRO_USER_V2)
    client.register(subject, version_2)

    compatibility = client.test_compatibility(subject, avro_user_schema_v3)
    assert compatibility


def test_avro_compatibility_dataclasses_avroschema(client, dataclass_avro_schema, dataclass_avro_schema_advance):
    """
    Test the compatibility of a new User Schema against the User schema version 2.
    """
    subject = "dataclasses-avroschema-subject"
    client.register(subject, dataclass_avro_schema.avro_schema())

    compatibility = client.test_compatibility(subject, dataclass_avro_schema_advance.avro_schema())
    assert compatibility


def test_avro_update_compatibility_for_subject(client):
    """
    The latest User V2 schema is  BACKWARD and FORWARDFULL compatibility (FULL).
    So, we can ipdate compatibility level for the specified subject.
    """
    assert client.update_compatibility("FULL", "test-avro-user-schema")


def test_avro_update_global_compatibility(client):
    """
    The latest User V2 schema is  BACKWARD and FORWARDFULL compatibility (FULL).
    So, we can ipdate compatibility level for the specified subject.
    """
    assert client.update_compatibility("FULL")


def test_avro_update_compatibility_fail(client, response_klass, mocker):
    http_code = 404
    mocker.patch.object(httpx.Client, "request", return_value=response_klass(http_code))

    with pytest.raises(errors.ClientError) as excinfo:
        client.update_compatibility("FULL", "test-avro-user-schema")

        assert excinfo.http_code == http_code


def test_avro_get_compatibility_for_subject(client):
    """
    Test latest compatibility for test-avro-user-schema subject
    """
    assert client.get_compatibility("test-avro-user-schema") == "FULL"


def test_avro_get_global_compatibility(client):
    """
    Test latest compatibility for test-avro-user-schema subject
    """
    assert client.get_compatibility() is not None


def test_json_compatibility(client, json_user_schema_v3):
    """
    Test the compatibility of a new User Schema against the User schema version 2.
    """
    subject = "test-json-user-schema"
    version_2 = schema.JsonSchema(data_gen.JSON_USER_V2)
    client.register(subject, version_2)

    compatibility = client.test_compatibility(subject, json_user_schema_v3)

    assert compatibility


def test_json_compatibility_dataclasses_jsonschema(client, dataclass_json_schema, dataclass_json_schema_advance):
    """
    Test the compatibility of a new User Schema against the User schema version 2.
    """
    subject = "dataclasses-jsonschema-subject"
    client.register(subject, dataclass_json_schema.schema_json(), schema_type=utils.JSON_SCHEMA_TYPE)

    compatibility = client.test_compatibility(
        subject, dataclass_json_schema_advance.schema_json(), schema_type=utils.JSON_SCHEMA_TYPE
    )

    assert compatibility


def test_json_update_compatibility_for_subject(client):
    """
    The latest User V2 schema is  BACKWARD and FORWARDFULL compatibility (FULL).
    So, we can ipdate compatibility level for the specified subject.
    """
    assert client.update_compatibility("FULL", "test-json-user-schema")


def test_json_update_global_compatibility(client):
    """
    The latest User V2 schema is  BACKWARD and FORWARDFULL compatibility (FULL).
    So, we can ipdate compatibility level for the specified subject.
    """
    assert client.update_compatibility("FULL")


def test_json_update_compatibility_fail(client, response_klass, mocker):
    http_code = 404
    mocker.patch.object(httpx.Client, "request", return_value=response_klass(http_code))

    with pytest.raises(errors.ClientError) as excinfo:
        client.update_compatibility("FULL", "test-json-user-schema")

        assert excinfo.http_code == http_code


def test_json_get_compatibility_for_subject(client):
    """
    Test latest compatibility for test-json-user-schema subject
    """
    assert client.get_compatibility("test-json-user-schema") == "FULL"


def test_json_get_global_compatibility(client):
    """
    Test latest compatibility for test-json-user-schema subject
    """
    assert client.get_compatibility() is not None



from schema_registry.client import schema as schema_loader
from tests import data_gen


def test_avro_getters(client):
    subject = "test-avro-basic-schema"
    parsed_basic = schema_loader.AvroSchema(data_gen.AVRO_BASIC_SCHEMA)
    client.register(subject, parsed_basic)
    schema = client.get_by_id(1)
    assert schema is not None

    subject = "subject-does-not-exist"
    latest = client.get_schema(subject)
    assert latest is None

    schema_id = client.register(subject, parsed_basic)
    latest = client.get_schema(subject)
    fetched = client.get_by_id(schema_id)

    assert fetched == parsed_basic


def test_avro_get_subjects(client, avro_user_schema_v3, avro_country_schema):
    subject_user = "test-avro-user-schema"
    subject_country = "test-avro-country"

    client.register("test-avro-user-schema", avro_user_schema_v3)
    client.register("test-avro-country", avro_country_schema)

    subjects = client.get_subjects()

    assert subject_user in subjects
    assert subject_country in subjects


def test_json_getters(client):
    subject = "test-json-basic-schema"
    parsed_basic = schema_loader.JsonSchema(data_gen.JSON_BASIC_SCHEMA)
    client.register(subject, parsed_basic)
    schema = client.get_by_id(1)
    assert schema is not None

    subject = "subject-does-not-exist"
    latest = client.get_schema(subject)
    assert latest is None

    schema_id = client.register(subject, parsed_basic)
    latest = client.get_schema(subject)
    fetched = client.get_by_id(schema_id)

    assert fetched == parsed_basic


def test_json_get_subjects(client, json_user_schema_v3, json_country_schema):
    subject_user = "test-json-user-schema"
    subject_country = "test-json-country"

    client.register("test-json-user-schema", json_user_schema_v3)
    client.register("test-json-country", json_country_schema)

    subjects = client.get_subjects()

    assert subject_user in subjects
    assert subject_country in subjects



from schema_registry.client import schema, utils
from tests import data_gen
from tests.conftest import RequestLoggingSchemaRegistryClient


def assertLatest(self, meta_tuple, sid, schema, version):
    self.assertNotEqual(sid, -1)
    self.assertNotEqual(version, -1)
    self.assertEqual(meta_tuple[0], sid)
    self.assertEqual(meta_tuple[1], schema)
    self.assertEqual(meta_tuple[2], version)


def test_avro_register(client):
    parsed = schema.AvroSchema(data_gen.AVRO_BASIC_SCHEMA)
    schema_id = client.register("test-avro-basic-schema", parsed)

    assert schema_id > 0
    assert len(client.id_to_schema) == 1

    schema_versions = client.get_schema_subject_versions(schema_id)
    assert len(schema_versions) == 1
    assert schema_versions[0].subject == "test-avro-basic-schema"


def test_avro_register_json_data(client, avro_deployment_schema):
    schema_id = client.register("test-avro-deployment", avro_deployment_schema)
    assert schema_id > 0


def test_avro_register_with_custom_headers(client, avro_country_schema):
    headers = {"custom-serialization": "application/x-avro-json"}
    schema_id = client.register("test-avro-country", avro_country_schema, headers=headers)
    assert schema_id > 0


def test_avro_register_with_logical_types(client):
    parsed = schema.AvroSchema(data_gen.AVRO_LOGICAL_TYPES_SCHEMA)
    schema_id = client.register("test-logical-types-schema", parsed)

    assert schema_id > 0
    assert len(client.id_to_schema) == 1


def test_avro_multi_subject_register(client: RequestLoggingSchemaRegistryClient):
    parsed = schema.AvroSchema(data_gen.AVRO_BASIC_SCHEMA)
    schema_id = client.register("test-avro-basic-schema", parsed)
    assert schema_id > 0

    # register again under different subject
    dupe_id = client.register("test-avro-basic-schema-backup", parsed)
    assert schema_id == dupe_id
    assert len(client.id_to_schema) == 1

    schema_versions = client.get_schema_subject_versions(schema_id)
    assert len(schema_versions) == 2
    schema_versions.sort(key=lambda x: x.subject)
    assert schema_versions[0].subject == "test-avro-basic-schema"
    assert schema_versions[1].subject == "test-avro-basic-schema-backup"
    # The schema version we get here has a tendency to vary with the
    # number of times the schema has been soft-deleted, so only verifying
    # it's an int and > 0
    assert type(schema_versions[1].version) == int
    assert schema_versions[1].version > 0


def test_avro_dupe_register(client):
    parsed = schema.AvroSchema(data_gen.AVRO_BASIC_SCHEMA)
    subject = "test-avro-basic-schema"
    schema_id = client.register(subject, parsed)

    # Verify we had a check version call
    client.assert_url_suffix(0, "/subjects/%s" % subject)
    client.assert_method(0, "POST")
    # Verify that we had a register call
    client.assert_url_suffix(1, "/subjects/%s/versions" % subject)
    client.assert_method(1, "POST")
    assert len(client.request_calls) == 2

    assert schema_id > 0
    latest = client.get_schema(subject)

    client.assert_url_suffix(2, "/subjects/%s/versions/latest" % subject)
    client.assert_method(2, "GET")
    assert len(client.request_calls) == 3

    # register again under same subject
    dupe_id = client.register(subject, parsed)
    assert schema_id == dupe_id

    # Served from cache
    assert len(client.request_calls) == 3

    dupe_latest = client.get_schema(subject)
    assert latest == dupe_latest


def test_avro_multi_register(client):
    """
    Register two different schemas under the same subject
    with backwards compatibility
    """
    version_1 = schema.AvroSchema(data_gen.AVRO_USER_V1)
    version_2 = schema.AvroSchema(data_gen.AVRO_USER_V2)
    subject = "test-avro-user-schema"

    id1 = client.register(subject, version_1)
    latest_schema_1 = client.get_schema(subject)
    client.check_version(subject, version_1)

    id2 = client.register(subject, version_2)
    latest_schema_2 = client.get_schema(subject)
    client.check_version(subject, version_2)

    assert id1 != id2
    assert latest_schema_1 != latest_schema_2
    # ensure version is higher
    assert latest_schema_1.version < latest_schema_2.version

    client.register(subject, version_1)
    latest_schema_3 = client.get_schema(subject)

    assert latest_schema_2 == latest_schema_3


def test_register_dataclass_avro_schema(client, dataclass_avro_schema):
    subject = "dataclasses-avroschema-subject"
    schema_id = client.register(subject, dataclass_avro_schema.avro_schema())

    assert schema_id > 0
    assert len(client.id_to_schema) == 1

    subjects = client.get_subjects()

    assert subject in subjects


def test_json_register(client):
    parsed = schema.JsonSchema(data_gen.JSON_BASIC_SCHEMA)
    schema_id = client.register("test-json-basic-schema", parsed)

    assert schema_id > 0
    assert len(client.id_to_schema) == 1


def test_json_register_json_data(client, json_deployment_schema):
    schema_id = client.register("test-json-deployment", json_deployment_schema)
    assert schema_id > 0


def test_json_register_with_custom_headers(client, json_country_schema):
    headers = {"custom-serialization": "application/x-avro-json"}
    schema_id = client.register("test-json-country", json_country_schema, headers=headers)
    assert schema_id > 0


def test_json_multi_subject_register(client):
    parsed = schema.JsonSchema(data_gen.JSON_BASIC_SCHEMA)
    schema_id = client.register("test-json-basic-schema", parsed)
    assert schema_id > 0

    # register again under different subject
    dupe_id = client.register("test-json-basic-schema-backup", parsed)
    assert schema_id == dupe_id
    assert len(client.id_to_schema) == 1


def test_json_dupe_register(client):
    parsed = schema.JsonSchema(data_gen.JSON_BASIC_SCHEMA)
    subject = "test-json-basic-schema"
    schema_id = client.register(subject, parsed)

    # Verify we had a check version call
    client.assert_url_suffix(0, "/subjects/%s" % subject)
    client.assert_method(0, "POST")
    # Verify that we had a register call
    client.assert_url_suffix(1, "/subjects/%s/versions" % subject)
    client.assert_method(1, "POST")
    assert len(client.request_calls) == 2

    assert schema_id > 0
    latest = client.get_schema(subject)

    client.assert_url_suffix(2, "/subjects/%s/versions/latest" % subject)
    client.assert_method(2, "GET")
    assert len(client.request_calls) == 3

    # register again under same subject
    dupe_id = client.register(subject, parsed)
    assert schema_id == dupe_id

    # Served from cache
    assert len(client.request_calls) == 3

    dupe_latest = client.get_schema(subject)
    assert latest == dupe_latest


def test_json_multi_register(client, json_user_schema_v3):
    """
    Register two different schemas under the same subject
    with backwards compatibility
    """
    version_1 = schema.JsonSchema(data_gen.JSON_USER_V2)
    version_2 = json_user_schema_v3
    subject = "test-json-user-schema"

    id1 = client.register(subject, version_1)
    latest_schema_1 = client.get_schema(subject)
    client.check_version(subject, version_1)

    id2 = client.register(subject, version_2)
    latest_schema_2 = client.get_schema(subject)
    client.check_version(subject, version_2)

    assert id1 != id2
    assert latest_schema_1 != latest_schema_2
    # ensure version is higher
    assert latest_schema_1.version < latest_schema_2.version

    client.register(subject, version_1)
    latest_schema_3 = client.get_schema(subject)

    assert latest_schema_2 == latest_schema_3


def test_register_dataclass_json_schema(client, dataclass_json_schema):
    subject = "dataclasses-jsonschema-subject"
    schema_id = client.register(subject, dataclass_json_schema.schema_json(), schema_type=utils.JSON_SCHEMA_TYPE)

    assert schema_id > 0
    assert len(client.id_to_schema) == 1

    subjects = client.get_subjects()

    assert subject in subjects



from collections import namedtuple

SchemaVersion = namedtuple("SchemaVersion", "subject schema_id schema version")

BACKWARD = "BACKWARD"
BACKWARD_TRANSITIVE = "BACKWARD_TRANSITIVE"
FORWARD = "FORWARD"
FORWARD_TRANSITIVE = "FORWARD_TRANSITIVE"
FULL = "FULL"
FULL_TRANSITIVE = "FULL_TRANSITIVE"
NONE = "NONE"

VALID_LEVELS = (BACKWARD, BACKWARD_TRANSITIVE, FORWARD, FORWARD_TRANSITIVE, FULL, FULL_TRANSITIVE, NONE)
VALID_METHODS = ("GET", "POST", "PUT", "DELETE")
VALID_AUTH_PROVIDERS = ("URL", "USER_INFO", "SASL_INHERIT")

HEADER_AVRO_JSON = "application/x-avro-json"
HEADER_AVRO = "application/avro"
HEADER_APPLICATION_JSON = "application/json"
HEADERS = "application/vnd.schemaregistry.v1+json"
ACCEPT_HEADERS = "application/vnd.schemaregistry.v1+json, application/vnd.schemaregistry+json, application/json"

URL = "url"
SSL_CA_LOCATION = "ssl.ca.location"
SSL_CERTIFICATE_LOCATION = "ssl.certificate.location"
SSL_KEY_LOCATION = "ssl.key.location"
SSL_KEY_PASSWORD = "ssl.key.password"

AVRO_SCHEMA_TYPE = "AVRO"
JSON_SCHEMA_TYPE = "JSON"



# Available urls supported with the client

paths = [
    ("get_subjects", "subjects", "GET"),
    ("get_versions", "subjects/{subject}/versions", "GET"),
    ("delete_version", "subjects/{subject}/versions/{version}", "DELETE"),
    ("register", "subjects/{subject}/versions", "POST"),
    ("delete_subject", "subjects/{subject}", "DELETE"),
    ("get_schema", "subjects/{subject}/versions/{version}", "GET"),
    ("check_version", "subjects/{subject}", "POST"),
    ("get_by_id", "schemas/ids/{schema_id}", "GET"),
    ("get_schema_subject_versions", "schemas/ids/{schema_id}/versions", "GET"),
    ("test_compatibility", "compatibility/subjects/{subject}/versions/{version}", "POST"),
    ("update_compatibility", "config/{subject}", "PUT"),
    ("get_compatibility", "config/{subject}", "GET"),
]



import urllib

import pytest

from schema_registry.client import urls
from schema_registry.client.paths import paths

BASE_URLS = ("http://localhost:8081", "http://localhost:8082/api/schema-registry/")


def test_fail_url_manager_creation():
    base_url = "localhost:8081"

    with pytest.raises(AssertionError):
        urls.UrlManager(base_url, [])


@pytest.mark.parametrize("base_url", BASE_URLS)
def test_url_with_path(base_url):
    paths = [("get_cars", "cars/{car_id}", "GET"), ("create_car", "cars", "POST")]

    url_manager = urls.UrlManager(base_url, paths)
    url, method = url_manager.url_for("get_cars")

    assert base_url in url


@pytest.mark.parametrize("base_url", BASE_URLS)
def test_urls_generation(base_url):
    local_paths = [("get_cars", "cars/{car_id}", "GET"), ("create_car", "cars", "POST")]

    url_manager = urls.UrlManager(base_url, local_paths)
    url, method = url_manager.url_for("get_cars")

    assert url == urllib.parse.urljoin(base_url, "cars/")
    assert method == "GET"

    url, method = url_manager.url_for("get_cars", car_id=10)
    assert url == urllib.parse.urljoin(base_url, "cars/10")
    assert method == "GET"

    url, method = url_manager.url_for("create_car")
    assert url == urllib.parse.urljoin(base_url, "cars")
    assert method == "POST"


@pytest.mark.parametrize("base_url", BASE_URLS)
def test_client_paths(base_url):
    url_manager = urls.UrlManager(base_url, paths)

    for func, path, _ in paths:
        kwargs = {"subject": "my-subject", "version": 1}
        url, method = url_manager.url_for(func, **kwargs)

        assert base_url in url



import pytest

from schema_registry.client import utils


@pytest.mark.asyncio
async def test_avro_version_does_not_exists(async_client, avro_country_schema):
    assert await async_client.check_version("test-avro-schema-version", avro_country_schema) is None


@pytest.mark.asyncio
async def test_avro_get_versions(async_client, avro_country_schema):
    subject = "test-avro-schema-version"
    await async_client.register(subject, avro_country_schema)
    versions = await async_client.get_versions(subject)

    assert versions


@pytest.mark.asyncio
async def test_avro_get_versions_does_not_exist(async_client):
    assert not await async_client.get_versions("random-subject")


@pytest.mark.asyncio
async def test_avro_check_version(async_client, avro_country_schema):
    subject = "test-avro-schema-version"
    schema_id = await async_client.register(subject, avro_country_schema)
    result = await async_client.check_version(subject, avro_country_schema)

    assert subject == result.subject
    assert schema_id == result.schema_id


@pytest.mark.asyncio
async def test_avro_check_version_dataclasses_avroschema(async_client, dataclass_avro_schema):
    subject = "dataclasses-avroschema-subject"
    schema_id = await async_client.register(subject, dataclass_avro_schema.avro_schema())
    result = await async_client.check_version(subject, dataclass_avro_schema.avro_schema())

    assert subject == result.subject
    assert schema_id == result.schema_id


@pytest.mark.asyncio
async def test_avro_delete_version(async_client, avro_country_schema):
    subject = "test-avro-schema-version"
    await async_client.register(subject, avro_country_schema)
    versions = await async_client.get_versions(subject)
    latest_version = versions[-1]

    assert latest_version == await async_client.delete_version(subject, latest_version)


@pytest.mark.asyncio
async def test_avro_delete_version_does_not_exist(async_client, avro_country_schema):
    subject = "test-avro-schema-version"
    await async_client.register(subject, avro_country_schema)

    assert not await async_client.delete_version("random-subject")
    assert not await async_client.delete_version(subject, "random-version")


@pytest.mark.asyncio
async def test_json_version_does_not_exists(async_client, json_country_schema):
    assert await async_client.check_version("test-json-schema-version", json_country_schema) is None


@pytest.mark.asyncio
async def test_json_get_versions(async_client, json_country_schema):
    subject = "test-json-schema-version"
    await async_client.register(subject, json_country_schema)
    versions = await async_client.get_versions(subject)

    assert versions


@pytest.mark.asyncio
async def test_json_get_versions_does_not_exist(async_client):
    assert not await async_client.get_versions("random-subject")


@pytest.mark.asyncio
async def test_json_check_version(async_client, json_country_schema):
    subject = "test-json-schema-version"
    schema_id = await async_client.register(subject, json_country_schema)
    result = await async_client.check_version(subject, json_country_schema)

    assert subject == result.subject
    assert schema_id == result.schema_id


@pytest.mark.asyncio
async def test_json_check_version_dataclasses_avroschema(async_client, dataclass_json_schema):
    subject = "dataclasses-jsonschema-subject"
    schema_id = await async_client.register(
        subject, dataclass_json_schema.schema_json(), schema_type=utils.JSON_SCHEMA_TYPE
    )
    result = await async_client.check_version(
        subject, dataclass_json_schema.schema_json(), schema_type=utils.JSON_SCHEMA_TYPE
    )

    assert subject == result.subject
    assert schema_id == result.schema_id


@pytest.mark.asyncio
async def test_json_delete_version(async_client, json_country_schema):
    subject = "test-json-schema-version"
    await async_client.register(subject, json_country_schema)
    versions = await async_client.get_versions(subject)
    latest_version = versions[-1]

    assert latest_version == await async_client.delete_version(subject, latest_version)


@pytest.mark.asyncio
async def test_json_delete_version_does_not_exist(async_client, json_country_schema):
    subject = "test-json-schema-version"
    await async_client.register(subject, json_country_schema)

    assert not await async_client.delete_version("random-subject")
    assert not await async_client.delete_version(subject, "random-version")



import typing

from faust import Record

from schema_registry.serializers import faust as serializer


class DummyRecord(Record):
    item: typing.Any


def test_avro_simple_record(client, avro_country_schema):
    schema_subject = "test-avro-country"
    faust_serializer = serializer.FaustSerializer(client, schema_subject, avro_country_schema)

    result = {"__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"}, "item": "test"}

    dummy = DummyRecord("test")
    assert result == faust_serializer.clean_payload(dummy)


def test_avro_nested_record(client, avro_country_schema):
    schema_subject = "test-avro-country"
    faust_serializer = serializer.FaustSerializer(client, schema_subject, avro_country_schema)

    result = {
        "__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"},
        "item": {"__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"}, "item": "test"},
    }

    dummy = DummyRecord(DummyRecord("test"))
    assert result == faust_serializer.clean_payload(dummy)


def test_avro_list_of_records(client, avro_country_schema):
    schema_subject = "test-avro-country"
    faust_serializer = serializer.FaustSerializer(client, schema_subject, avro_country_schema)

    result = {
        "__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"},
        "item": [
            {"__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"}, "item": "test"},
            {"__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"}, "item": "test"},
        ],
    }

    dummy = DummyRecord([DummyRecord("test"), DummyRecord("test")])
    assert result == faust_serializer.clean_payload(dummy)


def test_avro_map_of_records(client, avro_country_schema):
    schema_subject = "test-avro-country"
    faust_serializer = serializer.FaustSerializer(client, schema_subject, avro_country_schema)

    result = {
        "__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"},
        "item": {
            "key1": {
                "__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"},
                "item": "test",
            },
            "key2": {
                "__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"},
                "item": "test",
            },
        },
    }

    dummy = DummyRecord({"key1": DummyRecord("test"), "key2": DummyRecord("test")})
    assert result == faust_serializer.clean_payload(dummy)


def test_json_simple_record(client, json_country_schema):
    schema_subject = "test-avro-country"
    faust_serializer = serializer.FaustJsonSerializer(client, schema_subject, json_country_schema)

    result = {"__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"}, "item": "test"}

    dummy = DummyRecord("test")
    assert result == faust_serializer.clean_payload(dummy)


def test_json_nested_record(client, json_country_schema):
    schema_subject = "test-avro-country"
    faust_serializer = serializer.FaustJsonSerializer(client, schema_subject, json_country_schema)

    result = {
        "__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"},
        "item": {"__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"}, "item": "test"},
    }

    dummy = DummyRecord(DummyRecord("test"))
    assert result == faust_serializer.clean_payload(dummy)


def test_json_list_of_records(client, json_country_schema):
    schema_subject = "test-avro-country"
    faust_serializer = serializer.FaustJsonSerializer(client, schema_subject, json_country_schema)

    result = {
        "__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"},
        "item": [
            {"__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"}, "item": "test"},
            {"__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"}, "item": "test"},
        ],
    }

    dummy = DummyRecord([DummyRecord("test"), DummyRecord("test")])
    assert result == faust_serializer.clean_payload(dummy)


def test_json_map_of_records(client, json_country_schema):
    schema_subject = "test-avro-country"
    faust_serializer = serializer.FaustJsonSerializer(client, schema_subject, json_country_schema)

    result = {
        "__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"},
        "item": {
            "key1": {
                "__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"},
                "item": "test",
            },
            "key2": {
                "__faust": {"ns": "tests.serializer.test_faust_serializer_clean_payload.DummyRecord"},
                "item": "test",
            },
        },
    }

    dummy = DummyRecord({"key1": DummyRecord("test"), "key2": DummyRecord("test")})
    assert result == faust_serializer.clean_payload(dummy)



# import pickle
import os
from base64 import b64encode

import httpx
import pytest

from schema_registry.client import AsyncSchemaRegistryClient, utils


def test_invalid_cert():
    with pytest.raises(FileNotFoundError):
        AsyncSchemaRegistryClient(url="https://127.0.0.1:65534", cert_location="/path/to/cert")


def test_cert_with_key(certificates):
    client = AsyncSchemaRegistryClient(
        url="https://127.0.0.1:65534",
        cert_location=certificates["certificate"],
        key_location=certificates["key"],
        key_password=certificates["password"],
    )

    assert client.conf[utils.SSL_CERTIFICATE_LOCATION] == certificates["certificate"]
    assert client.conf[utils.SSL_KEY_LOCATION] == certificates["key"]
    assert client.conf[utils.SSL_KEY_PASSWORD] == certificates["password"]


def test_custom_headers():
    extra_headers = {"custom-serialization": utils.HEADER_AVRO_JSON}

    client = AsyncSchemaRegistryClient(url="https://127.0.0.1:65534", extra_headers=extra_headers)
    assert extra_headers == client.extra_headers


@pytest.mark.asyncio
async def test_override_headers(avro_deployment_schema, response_klass, async_mock):
    extra_headers = {"custom-serialization": utils.HEADER_AVRO_JSON}
    async_client = AsyncSchemaRegistryClient(url=os.getenv("SCHEMA_REGISTRY_URL"), extra_headers=extra_headers)

    assert async_client.session.headers.get("custom-serialization") == utils.HEADER_AVRO_JSON

    subject = "test"
    override_header = {"custom-serialization": utils.HEADER_AVRO}

    mock = async_mock(httpx.AsyncClient, "request", returned_value=response_klass(200, content={"id": 1}))

    with mock:
        await async_client.register(subject, avro_deployment_schema, headers=override_header)

        prepare_headers = async_client.prepare_headers(body="1")
        prepare_headers["custom-serialization"] = utils.HEADER_AVRO

        mock.assert_called_with(headers=prepare_headers)


def test_cert_path():
    client = AsyncSchemaRegistryClient(url="https://127.0.0.1:65534", ca_location=True)

    assert client.conf[utils.SSL_CA_LOCATION]


def test_init_with_dict(certificates):
    client = AsyncSchemaRegistryClient(
        {
            "url": "https://127.0.0.1:65534",
            "ssl.certificate.location": certificates["certificate"],
            "ssl.key.location": certificates["key"],
            "ssl.key.password": "test",
        }
    )
    assert "https://127.0.0.1:65534/" == client.url_manager.url


def test_empty_url():
    with pytest.raises(AssertionError):
        AsyncSchemaRegistryClient({"url": ""})


def test_invalid_type_url():
    with pytest.raises(AttributeError):
        AsyncSchemaRegistryClient(url=1)


def test_invalid_type_url_dict():
    with pytest.raises(AttributeError):
        AsyncSchemaRegistryClient({"url": 1})


def test_basic_auth_url():
    username = "secret-user"
    password = "secret"
    client = AsyncSchemaRegistryClient({"url": f"https://{username}:{password}@127.0.0.1:65534"})
    userpass = b":".join((httpx._utils.to_bytes(username), httpx._utils.to_bytes(password)))
    token = b64encode(userpass).decode()

    assert client.session.auth._auth_header == f"Basic {token}"


def test_basic_auth_user_info():
    username = "secret-user"
    password = "secret"
    client = AsyncSchemaRegistryClient(
        {
            "url": "https://user_url:secret_url@127.0.0.1:65534",
            "basic.auth.credentials.source": "user_info",
            "basic.auth.user.info": f"{username}:{password}",
        }
    )

    userpass = b":".join((httpx._utils.to_bytes(username), httpx._utils.to_bytes(password)))
    token = b64encode(userpass).decode()

    assert client.session.auth._auth_header == f"Basic {token}"


def test_basic_auth_sasl_inherit():
    username = "secret-user-sasl"
    password = "secret-sasl"
    client = AsyncSchemaRegistryClient(
        {
            "url": "https://user_url:secret_url@127.0.0.1:65534",
            "basic.auth.credentials.source": "SASL_INHERIT",
            "sasl.mechanism": "PLAIN",
            "sasl.username": username,
            "sasl.password": password,
        }
    )

    userpass = b":".join((httpx._utils.to_bytes(username), httpx._utils.to_bytes(password)))
    token = b64encode(userpass).decode()

    assert client.session.auth._auth_header == f"Basic {token}"


def test_basic_auth_invalid():
    with pytest.raises(ValueError):
        AsyncSchemaRegistryClient(
            {"url": "https://user_url:secret_url@127.0.0.1:65534", "basic.auth.credentials.source": "VAULT"}
        )



import io
import json
import logging
import struct
import sys
import traceback
import typing
from abc import ABC, abstractmethod

from fastavro import schemaless_reader, schemaless_writer
from jsonschema import validate

from schema_registry.client import SchemaRegistryClient, schema, utils, AsyncSchemaRegistryClient
from schema_registry.client.errors import ClientError
from schema_registry.client.schema import BaseSchema

from .errors import SerializerError

log = logging.getLogger(__name__)

MAGIC_BYTE = 0


class ContextStringIO(io.BytesIO):
    """
    Wrapper to allow use of StringIO via 'with' constructs.
    """

    def __enter__(self) -> "ContextStringIO":
        return self

    def __exit__(self, *args: typing.Any) -> None:
        self.close()


class MessageSerializer(ABC):
    """
    A helper class that can serialize and deserialize messages
    Args:
        schemaregistry_client (schema_registry.client.SchemaRegistryClient): Http Client
    """

    def __init__(
        self,
        schemaregistry_client: SchemaRegistryClient,
        reader_schema: typing.Optional[schema.AvroSchema] = None,
        return_record_name: bool = False,
    ):
        self.schemaregistry_client = schemaregistry_client
        self.id_to_decoder_func = {}  # type: typing.Dict
        self.id_to_writers = {}  # type: typing.Dict
        self.reader_schema = reader_schema
        self.return_record_name = return_record_name

    @property
    @abstractmethod
    def _serializer_schema_type(self) -> str:
        ...

    @abstractmethod
    def _get_encoder_func(self, schema: BaseSchema) -> typing.Callable:
        ...

    @abstractmethod
    def _get_decoder_func(self, payload: ContextStringIO, writer_schema: BaseSchema) -> typing.Callable:
        ...

    def encode_record_with_schema(self, subject: str, schema: typing.Union[BaseSchema], record: dict) -> bytes:
        """
        Given a parsed avro schema, encode a record for the given subject.
        The record is expected to be a dictionary.
        The schema is registered with the subject of 'topic-value'
        Args:
            subject (str): Subject name
            schema (avro.schema.RecordSchema): Avro Schema
            record (dict): An object to serialize
        Returns:
            bytes: Encoded record with schema ID as bytes
        """
        # Try to register the schema
        schema_id = self.schemaregistry_client.register(subject, schema, schema_type=self._serializer_schema_type)

        # cache writer
        if not self.id_to_writers.get(schema_id):
            self.id_to_writers[schema_id] = self._get_encoder_func(schema)

        return self.encode_record_with_schema_id(schema_id, record)

    def encode_record_with_schema_id(self, schema_id: int, record: dict) -> bytes:
        """
        Encode a record with a given schema id.  The record must
        be a python dictionary.
        Args:
            schema_id (int): integer ID
            record (dict): An object to serialize
        Returns:
            func: decoder function
        """
        # use slow avro
        if schema_id not in self.id_to_writers:
            try:
                schema = self.schemaregistry_client.get_by_id(schema_id)
                if not schema:
                    raise SerializerError("Schema does not exist")
                self.id_to_writers[schema_id] = self._get_encoder_func(schema)
            except ClientError:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                raise SerializerError(repr(traceback.format_exception(exc_type, exc_value, exc_traceback)))

        writer = self.id_to_writers[schema_id]
        with ContextStringIO() as outf:
            # Write the magic byte and schema ID in network byte order (big endian)
            outf.write(struct.pack(">bI", MAGIC_BYTE, schema_id))

            # write the record to the rest of the buffer
            writer(record, outf)

            return outf.getvalue()

    def decode_message(self, message: typing.Optional[bytes]) -> typing.Optional[dict]:
        """
        Decode a message from kafka that has been encoded for use with
        the schema registry.
        Args:
            message (bytes or None): message key or value to be decoded
        Returns:
            dict: Decoded message contents.
        """

        if message is None:
            return None

        if len(message) <= 5:
            raise SerializerError("message is too small to decode")

        with ContextStringIO(message) as payload:
            magic, schema_id = struct.unpack(">bI", payload.read(5))
            if magic != MAGIC_BYTE:
                raise SerializerError("message does not start with magic byte")

            if schema_id in self.id_to_decoder_func:
                return self.id_to_decoder_func[schema_id](payload)

            try:
                writer_schema = self.schemaregistry_client.get_by_id(schema_id)
            except ClientError as e:
                raise SerializerError(f"unable to fetch schema with id {schema_id}: {e}")

            if writer_schema is None:
                raise SerializerError(f"unable to fetch schema with id {schema_id}")

            decoder_func = self._get_decoder_func(payload, writer_schema)
            self.id_to_decoder_func[schema_id] = decoder_func

            return decoder_func(payload)


class AvroMessageSerializer(MessageSerializer):
    @property
    def _serializer_schema_type(self) -> str:
        return utils.AVRO_SCHEMA_TYPE

    def _get_encoder_func(self, schema: typing.Union[BaseSchema]) -> typing.Callable:
        return lambda record, fp: schemaless_writer(fp, schema.schema, record)  # type: ignore

    def _get_decoder_func(self, payload: ContextStringIO, writer_schema: BaseSchema) -> typing.Callable:
        return lambda payload: schemaless_reader(
            payload, writer_schema.schema, self.reader_schema, self.return_record_name
        )  # type: ignore


class JsonMessageSerializer(MessageSerializer):
    @property
    def _serializer_schema_type(self) -> str:
        return utils.JSON_SCHEMA_TYPE

    def _get_encoder_func(self, schema: typing.Union[BaseSchema]) -> typing.Callable:
        def json_encoder_func(record: dict, fp: ContextStringIO) -> typing.Any:
            validate(record, schema.schema)
            fp.write(json.dumps(record).encode())

        return json_encoder_func

    def _get_decoder_func(self, payload: ContextStringIO, writer_schema: BaseSchema) -> typing.Callable:
        def json_decoder_func(payload: typing.Union[str, bytes]) -> typing.Any:
            obj = json.load(payload)  # type: ignore
            validate(obj, writer_schema.schema)  # type: ignore
            return obj

        return json_decoder_func


class AsyncMessageSerializer(ABC):
    """
    A helper class that can serialize and deserialize messages asynchronously
    Args:
        schemaregistry_client (schema_registry.client.AsyncSchemaRegistryClient): Http Client
        reader_schema (schema_registry.schema.AvroSchema): Specify a schema to decode the message
        return_record_name (bool): If the record name should be returned
    """

    def __init__(
        self,
        schemaregistry_client: AsyncSchemaRegistryClient,
        reader_schema: typing.Optional[schema.AvroSchema] = None,
        return_record_name: bool = False,
    ):
        self.schemaregistry_client = schemaregistry_client
        self.id_to_decoder_func = {}  # type: typing.Dict
        self.id_to_writers = {}  # type: typing.Dict
        self.reader_schema = reader_schema
        self.return_record_name = return_record_name

    @property
    @abstractmethod
    def _serializer_schema_type(self) -> str:
        ...

    @abstractmethod
    def _get_encoder_func(self, schema: BaseSchema) -> typing.Callable:
        ...

    @abstractmethod
    def _get_decoder_func(self, payload: ContextStringIO, writer_schema: BaseSchema) -> typing.Callable:
        ...

    async def encode_record_with_schema(self, subject: str, schema: typing.Union[BaseSchema], record: dict) -> bytes:
        """
        Given a parsed avro schema, encode a record for the given subject.
        The record is expected to be a dictionary.
        The schema is registered with the subject of 'topic-value'
        Args:
            subject (str): Subject name
            schema (avro.schema.RecordSchema): Avro Schema
            record (dict): An object to serialize
        Returns:
            bytes: Encoded record with schema ID as bytes
        """
        # Try to register the schema
        schema_id = await self.schemaregistry_client.register(subject, schema, schema_type=self._serializer_schema_type)

        # cache writer
        if not self.id_to_writers.get(schema_id):
            self.id_to_writers[schema_id] = self._get_encoder_func(schema)

        return await self.encode_record_with_schema_id(schema_id, record)

    async def encode_record_with_schema_id(self, schema_id: int, record: dict) -> bytes:
        """
        Encode a record with a given schema id.  The record must
        be a python dictionary.
        Args:
            schema_id (int): integer ID
            record (dict): An object to serialize
        Returns:
            func: decoder function
        """
        # use slow avro
        if schema_id not in self.id_to_writers:
            try:
                schema = await self.schemaregistry_client.get_by_id(schema_id)
                if not schema:
                    raise SerializerError("Schema does not exist")
                self.id_to_writers[schema_id] = self._get_encoder_func(schema)
            except ClientError:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                raise SerializerError(repr(traceback.format_exception(exc_type, exc_value, exc_traceback)))

        writer = self.id_to_writers[schema_id]
        with ContextStringIO() as outf:
            # Write the magic byte and schema ID in network byte order (big endian)
            outf.write(struct.pack(">bI", MAGIC_BYTE, schema_id))

            # write the record to the rest of the buffer
            writer(record, outf)

            return outf.getvalue()

    async def decode_message(self, message: typing.Optional[bytes]) -> typing.Optional[dict]:
        """
        Decode a message from kafka that has been encoded for use with
        the schema registry.
        Args:
            message (bytes or None): message key or value to be decoded
        Returns:
            dict: Decoded message contents.
        """

        if message is None:
            return None

        if len(message) <= 5:
            raise SerializerError("message is too small to decode")

        with ContextStringIO(message) as payload:
            magic, schema_id = struct.unpack(">bI", payload.read(5))
            if magic != MAGIC_BYTE:
                raise SerializerError("message does not start with magic byte")

            if schema_id in self.id_to_decoder_func:
                return self.id_to_decoder_func[schema_id](payload)

            try:
                writer_schema = await self.schemaregistry_client.get_by_id(schema_id)
            except ClientError as e:
                raise SerializerError(f"unable to fetch schema with id {schema_id}: {e}")

            if writer_schema is None:
                raise SerializerError(f"unable to fetch schema with id {schema_id}")

            decoder_func = self._get_decoder_func(payload, writer_schema)
            self.id_to_decoder_func[schema_id] = decoder_func

            return decoder_func(payload)


class AsyncAvroMessageSerializer(AsyncMessageSerializer):
    @property
    def _serializer_schema_type(self) -> str:
        return utils.AVRO_SCHEMA_TYPE

    def _get_encoder_func(self, schema: typing.Union[BaseSchema]) -> typing.Callable:
        return lambda record, fp: schemaless_writer(fp, schema.schema, record)  # type: ignore

    def _get_decoder_func(self, payload: ContextStringIO, writer_schema: BaseSchema) -> typing.Callable:
        return lambda payload: schemaless_reader(
            payload, writer_schema.schema, self.reader_schema, self.return_record_name
        )  # type: ignore


class AsyncJsonMessageSerializer(AsyncMessageSerializer):
    @property
    def _serializer_schema_type(self) -> str:
        return utils.JSON_SCHEMA_TYPE

    def _get_encoder_func(self, schema: typing.Union[BaseSchema]) -> typing.Callable:
        def json_encoder_func(record: dict, fp: ContextStringIO) -> typing.Any:
            validate(record, schema.schema)
            fp.write(json.dumps(record).encode())

        return json_encoder_func

    def _get_decoder_func(self, payload: ContextStringIO, writer_schema: BaseSchema) -> typing.Callable:
        def json_decoder_func(payload: typing.Union[str, bytes]) -> typing.Any:
            obj = json.load(payload)  # type: ignore
            validate(obj, writer_schema.schema)  # type: ignore
            return obj

        return json_decoder_func



from schema_registry.serializers.message_serializer import AvroMessageSerializer  # noqa
from schema_registry.serializers.message_serializer import JsonMessageSerializer  # noqa
from schema_registry.serializers.message_serializer import MessageSerializer  # noqa
from schema_registry.serializers.message_serializer import AsyncAvroMessageSerializer  # noqa
from schema_registry.serializers.message_serializer import AsyncJsonMessageSerializer  # noqa
from schema_registry.serializers.message_serializer import AsyncMessageSerializer  # noqa



import struct

import pytest
import jsonschema

from schema_registry.client import schema
from tests import data_gen


def assertAvroMessageIsSame(message, expected, schema_id, avro_message_serializer):
    assert message
    assert len(message) > 5

    magic, sid = struct.unpack(">bI", message[0:5])
    assert magic == 0
    assert sid == schema_id

    decoded = avro_message_serializer.decode_message(message)
    assert decoded
    assert decoded == expected


def test_avro_encode_with_schema_id(client, avro_message_serializer):
    basic = schema.AvroSchema(data_gen.AVRO_BASIC_SCHEMA)
    subject = "test-avro-basic-schema"
    schema_id = client.register(subject, basic)

    records = data_gen.AVRO_BASIC_ITEMS
    for record in records:
        message = avro_message_serializer.encode_record_with_schema_id(schema_id, record)
        assertAvroMessageIsSame(message, record, schema_id, avro_message_serializer)

    adv = schema.AvroSchema(data_gen.AVRO_ADVANCED_SCHEMA)
    subject = "test-avro-advance-schema"
    adv_schema_id = client.register(subject, adv)

    assert adv_schema_id != schema_id

    records = data_gen.AVRO_ADVANCED_ITEMS
    for record in records:
        message = avro_message_serializer.encode_record_with_schema_id(adv_schema_id, record)
        assertAvroMessageIsSame(message, record, adv_schema_id, avro_message_serializer)


def test_avro_encode_logical_types(client, avro_message_serializer):
    logical_types_schema = schema.AvroSchema(data_gen.AVRO_LOGICAL_TYPES_SCHEMA)
    subject = "test-logical-types-schema"
    schema_id = client.register(subject, logical_types_schema)

    record = data_gen.create_logical_item()
    message = avro_message_serializer.encode_record_with_schema_id(schema_id, record)

    decoded = avro_message_serializer.decode_message(message)

    decoded_datetime = decoded.get("metadata").get("timestamp")
    timestamp = record.get("metadata").get("timestamp")

    decoded_total = decoded.get("metadata").get("total")
    total = record.get("metadata").get("total")

    assert timestamp == decoded_datetime.replace(tzinfo=None)
    assert total == decoded_total


def test_avro_encode_decode_with_schema_from_json(avro_message_serializer, avro_deployment_schema):
    deployment_record = {"image": "registry.gitlab.com/my-project:1.0.0", "replicas": 1, "port": 8080}

    message_encoded = avro_message_serializer.encode_record_with_schema(
        "avro-deployment", avro_deployment_schema, deployment_record
    )

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    # now decode the message
    message_decoded = avro_message_serializer.decode_message(message_encoded)
    assert message_decoded == deployment_record


# def test_encode_with_schema_string(avro_message_serializer):
#     deployment_record = {"image": "registry.gitlab.com/my-project:1.0.0", "replicas": 1, "port": 8080}
#     schema = """{
#         "type": "record",
#         "namespace": "com.kubertenes.v2",
#         "name": "AvroDeploymentV2",
#         "fields": [
#             {"name": "image", "type": "string"},
#             {"name": "replicas", "type": "int"},
#             {"name": "host", "type": "string", "default": "localhost"},
#             {"name": "port", "type": "int"}
#         ]
#     }"""

#     message_encoded = avro_message_serializer.encode_record_with_schema(
#         "avro-deployment", schema, deployment_record
#     )

#     assert message_encoded
#     assert len(message_encoded) > 5
#     assert isinstance(message_encoded, bytes)

#     # now decode the message
#     message_decoded = avro_message_serializer.decode_message(message_encoded)
#     assert message_decoded == deployment_record


def test_avro_fail_encode_with_schema(avro_message_serializer, avro_deployment_schema):
    bad_record = {"image": "registry.gitlab.com/my-project:1.0.0", "replicas": "1", "port": "8080"}

    with pytest.raises(TypeError):
        avro_message_serializer.encode_record_with_schema("avro-deployment", avro_deployment_schema, bad_record)


def test_avro_encode_record_with_schema(client, avro_message_serializer):
    topic = "test"
    basic = schema.AvroSchema(data_gen.AVRO_BASIC_SCHEMA)
    subject = "test-avro-value"
    schema_id = client.register(subject, basic)
    records = data_gen.AVRO_BASIC_ITEMS

    for record in records:
        message = avro_message_serializer.encode_record_with_schema(topic, basic, record)
        assertAvroMessageIsSame(message, record, schema_id, avro_message_serializer)


def test_avro_decode_none(avro_message_serializer):
    """ "null/None messages should decode to None"""

    assert avro_message_serializer.decode_message(None) is None


def assertJsonMessageIsSame(message, expected, schema_id, json_message_serializer):
    assert message
    assert len(message) > 5

    magic, sid = struct.unpack(">bI", message[0:5])
    assert magic == 0
    assert sid == schema_id

    decoded = json_message_serializer.decode_message(message)
    assert decoded
    assert decoded == expected


def test_json_encode_with_schema_id(client, json_message_serializer):
    basic = schema.JsonSchema(data_gen.JSON_BASIC_SCHEMA)
    subject = "test-json-basic-schema"
    schema_id = client.register(subject, basic)

    records = data_gen.JSON_BASIC_ITEMS
    for record in records:
        message = json_message_serializer.encode_record_with_schema_id(schema_id, record)
        assertJsonMessageIsSame(message, record, schema_id, json_message_serializer)

    adv = schema.JsonSchema(data_gen.JSON_ADVANCED_SCHEMA)
    subject = "test-json-advance-schema"
    adv_schema_id = client.register(subject, adv)

    assert adv_schema_id != schema_id

    records = data_gen.JSON_ADVANCED_ITEMS
    for record in records:
        message = json_message_serializer.encode_record_with_schema_id(adv_schema_id, record)
        assertJsonMessageIsSame(message, record, adv_schema_id, json_message_serializer)


def test_json_encode_decode_with_schema_from_json(json_message_serializer, json_deployment_schema):
    deployment_record = {"image": "registry.gitlab.com/my-project:1.0.0", "replicas": 1, "port": 8080}

    message_encoded = json_message_serializer.encode_record_with_schema(
        "json-deployment", json_deployment_schema, deployment_record
    )

    assert message_encoded
    assert len(message_encoded) > 5
    assert isinstance(message_encoded, bytes)

    # now decode the message
    message_decoded = json_message_serializer.decode_message(message_encoded)
    assert message_decoded == deployment_record


def test_json_fail_encode_with_schema(json_message_serializer, json_deployment_schema):
    bad_record = {"image": "registry.gitlab.com/my-project:1.0.0", "replicas": "1", "port": "8080"}

    with pytest.raises(jsonschema.exceptions.ValidationError):
        json_message_serializer.encode_record_with_schema("json-deployment", json_deployment_schema, bad_record)


def test_json_encode_record_with_schema(client, json_message_serializer):
    topic = "test"
    basic = schema.JsonSchema(data_gen.JSON_BASIC_SCHEMA)
    subject = "test-json-value"
    schema_id = client.register(subject, basic)
    records = data_gen.JSON_BASIC_ITEMS

    for record in records:
        message = json_message_serializer.encode_record_with_schema(topic, basic, record)
        assertJsonMessageIsSame(message, record, schema_id, json_message_serializer)


def test_json_decode_none(json_message_serializer):
    """ "null/None messages should decode to None"""

    assert json_message_serializer.decode_message(None) is None



class ClientError(Exception):
    """Error thrown by Schema Registry client"""

    def __init__(self, message: str, http_code: int = None, server_traceback: str = None) -> None:
        self.message = message
        self.server_traceback = server_traceback
        self.http_code = http_code
        super().__init__(message)

    def __repr__(self) -> str:
        return f"ClientError(error={self.message})"

    def __str__(self) -> str:
        return self.message



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" setup.py for python-schema-registry-client."""

from setuptools import find_packages, setup

__version__ = "2.2.0"

with open("README.md") as readme_file:
    long_description = readme_file.read()


requires = ["fastavro>=1.4.4", "jsonschema>=3.2.0", "httpx>=0.14,<0.18", "aiofiles>=0.7.0", "dataclasses>=0.8; python_version < '3.7'"]

description = "Python Rest Client to interact against Schema Registry Confluent Server to manage Avro Schemas"

setup(
    name="python-schema-registry-client",
    version=__version__,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Marcos Schroh",
    author_email="schrohm@gmail.com",
    install_requires=requires,
    extras_require={
        "faust": ["faust-streaming",],
        "docs": ["mkdocs", "mkdocs-material",],
        "tests": [
            "black",
            "autoflake",
            "flake8",
            "mypy==0.782",
            "isort",
            "pytest",
            "pytest-mock",
            "pytest-asyncio",
            "faker",
            "codecov",
            "pytest-cov",
            "dataclasses-avroschema",
            "pydantic"
        ],
    },
    url="https://github.com/marcosschroh/python-schema-registry-client",
    download_url="https://pypi.org/project/python-schema-registry-client/#files",
    packages=find_packages(exclude=("tests", "docs",)),
    include_package_data=True,
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development",
    ],
    keywords=(
        """
        Schema Registry, Python, Avro, Apache, Apache Avro, JSON, JSON Schema
        """
    ),
)



import pickle
from base64 import b64encode

import httpx
import pytest
from httpx._client import UNSET

from schema_registry.client import SchemaRegistryClient, schema, utils
from tests import data_gen


def test_invalid_cert():
    with pytest.raises(FileNotFoundError):
        SchemaRegistryClient(url="https://127.0.0.1:65534", cert_location="/path/to/cert")


def test_cert_with_key(certificates):
    client = SchemaRegistryClient(
        url="https://127.0.0.1:65534",
        cert_location=certificates["certificate"],
        key_location=certificates["key"],
        key_password=certificates["password"],
    )

    assert client.conf[utils.SSL_CERTIFICATE_LOCATION] == certificates["certificate"]
    assert client.conf[utils.SSL_KEY_LOCATION] == certificates["key"]
    assert client.conf[utils.SSL_KEY_PASSWORD] == certificates["password"]


def test_pickelable(client):
    unpickled_client = pickle.loads(pickle.dumps(client))

    assert client == unpickled_client

    # make sure that is possible to do client operations with unpickled_client
    subject = "test-avro-basic-schema"
    parsed = schema.AvroSchema(data_gen.AVRO_BASIC_SCHEMA)
    unpickled_client.get_subjects()
    schema_id = unpickled_client.register(subject, parsed)

    assert schema_id > 0
    assert unpickled_client.delete_subject(subject)


def test_custom_headers():
    extra_headers = {"custom-serialization": utils.HEADER_AVRO_JSON}

    client = SchemaRegistryClient(url="https://127.0.0.1:65534", extra_headers=extra_headers)
    assert extra_headers == client.extra_headers


def test_custom_httpx_config():
    """
    Test the SchemaRegistryClient creation with custom httpx config
    """
    timeout = httpx.Timeout(10.0, connect=60.0)
    pool_limits = httpx.Limits(max_keepalive_connections=5, max_connections=10)

    client = SchemaRegistryClient(
        url="https://127.0.0.1:65534",
        timeout=timeout,
        pool_limits=pool_limits,
    )

    assert client.timeout == timeout
    assert client.pool_limits == pool_limits


def test_override_headers(client, avro_deployment_schema, mocker, response_klass):
    extra_headers = {"custom-serialization": utils.HEADER_AVRO_JSON}
    client = SchemaRegistryClient("https://127.0.0.1:65534", extra_headers=extra_headers)

    assert client.session.headers.get("custom-serialization") == utils.HEADER_AVRO_JSON

    subject = "test"
    override_header = {"custom-serialization": utils.HEADER_AVRO}

    request_patch = mocker.patch.object(httpx.Client, "request", return_value=response_klass(200, content={"id": 1}))
    client.register(subject, avro_deployment_schema, headers=override_header)

    prepare_headers = client.prepare_headers(body="1")
    prepare_headers["custom-serialization"] = utils.HEADER_AVRO

    request_patch.assert_called_once_with("POST", mocker.ANY, headers=prepare_headers, json=mocker.ANY, timeout=UNSET)


def test_cert_path():
    client = SchemaRegistryClient(url="https://127.0.0.1:65534", ca_location=True)

    assert client.conf[utils.SSL_CA_LOCATION]


def test_init_with_dict(certificates):
    client = SchemaRegistryClient(
        {
            "url": "https://127.0.0.1:65534",
            "ssl.certificate.location": certificates["certificate"],
            "ssl.key.location": certificates["key"],
            "ssl.key.password": certificates["password"],
        }
    )
    assert "https://127.0.0.1:65534/" == client.url_manager.url


def test_empty_url():
    with pytest.raises(AssertionError):
        SchemaRegistryClient({"url": ""})


def test_invalid_type_url():
    with pytest.raises(AttributeError):
        SchemaRegistryClient(url=1)


def test_invalid_type_url_dict():
    with pytest.raises(AttributeError):
        SchemaRegistryClient({"url": 1})


def test_invalid_url():
    with pytest.raises(AssertionError):
        SchemaRegistryClient({"url": "example.com:65534"})


def test_basic_auth_url():
    username = "secret-user"
    password = "secret"
    client = SchemaRegistryClient({"url": f"https://{username}:{password}@127.0.0.1:65534"})
    userpass = b":".join((httpx._utils.to_bytes(username), httpx._utils.to_bytes(password)))
    token = b64encode(userpass).decode()

    assert client.session.auth._auth_header == f"Basic {token}"


def test_basic_auth_user_info():
    username = "secret-user"
    password = "secret"
    client = SchemaRegistryClient(
        {
            "url": "https://user_url:secret_url@127.0.0.1:65534",
            "basic.auth.credentials.source": "user_info",
            "basic.auth.user.info": f"{username}:{password}",
        }
    )

    userpass = b":".join((httpx._utils.to_bytes(username), httpx._utils.to_bytes(password)))
    token = b64encode(userpass).decode()

    assert client.session.auth._auth_header == f"Basic {token}"


def test_basic_auth_sasl_inherit():
    username = "secret-user-sasl"
    password = "secret-sasl"
    client = SchemaRegistryClient(
        {
            "url": "https://user_url:secret_url@127.0.0.1:65534",
            "basic.auth.credentials.source": "SASL_INHERIT",
            "sasl.mechanism": "PLAIN",
            "sasl.username": username,
            "sasl.password": password,
        }
    )

    userpass = b":".join((httpx._utils.to_bytes(username), httpx._utils.to_bytes(password)))
    token = b64encode(userpass).decode()

    assert client.session.auth._auth_header == f"Basic {token}"


def test_basic_auth_invalid():
    with pytest.raises(ValueError):
        SchemaRegistryClient(
            {"url": "https://user_url:secret_url@127.0.0.1:65534", "basic.auth.credentials.source": "VAULT"}
        )
